#cython: boundscheck=False, wraparound=False, initializedcheck=False

cimport numpy as np
import numpy as np
cimport cython
from finesse.knm.matrix cimport make_unscaled_X_scatter_knm_matrix, make_unscaled_Y_scatter_knm_matrix
from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, cexp
from finesse.cymath.gaussbeam cimport bp_beamsize
from finesse.cymath.math cimport exp, sqrt, cos, sin, asin
from finesse.cymath.math cimport radians
from finesse.frequency cimport FrequencyContainer
from finesse.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.parameter cimport Parameter, ParameterState
from finesse.symbols import Symbol
from finesse.cymath.complex cimport DenseZVector
from finesse.utilities import refractive_index

from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.string cimport strcmp, memcpy
from libc.stdlib cimport free, calloc

import logging

ctypedef (double*, double*, double*, double*, double*, double*, double*, double*, double*) ptr_tuple_9

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD


LOGGER = logging.getLogger(__name__)


cdef class BeamsplitterOpticalConnections:
    def __cinit__(self, object bs, MatrixSystemSolver mtx):
        # Only 1D arrays of views as spaces don't
        # couple frequencies together.
        Nf = mtx.optical_frequencies.size
        self.P1i_P2o = SubCCSView1DArray(Nf)
        self.P2i_P1o = SubCCSView1DArray(Nf)
        self.P3i_P4o = SubCCSView1DArray(Nf)
        self.P4i_P3o = SubCCSView1DArray(Nf)
        self.P1i_P3o = SubCCSView1DArray(Nf)
        self.P3i_P1o = SubCCSView1DArray(Nf)
        self.P2i_P4o = SubCCSView1DArray(Nf)
        self.P4i_P2o = SubCCSView1DArray(Nf)

        self.opt_conn_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.opt_conn_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views
        self.opt_conn_ptrs.P3i_P4o = <PyObject**>self.P3i_P4o.views
        self.opt_conn_ptrs.P4i_P3o = <PyObject**>self.P4i_P3o.views
        self.opt_conn_ptrs.P1i_P3o = <PyObject**>self.P1i_P3o.views
        self.opt_conn_ptrs.P3i_P1o = <PyObject**>self.P3i_P1o.views
        self.opt_conn_ptrs.P2i_P4o = <PyObject**>self.P2i_P4o.views
        self.opt_conn_ptrs.P4i_P2o = <PyObject**>self.P4i_P2o.views

cdef class BeamsplitterSignalConnections(BeamsplitterOpticalConnections):
    def __cinit__(self, object bs, MatrixSystemSolver mtx):
        cdef:
            int Nfo =  mtx.optical_frequencies.size

        Nmz = bs.mech.z.num_frequencies # num of mechanic frequencies

        self.P1i_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P1o_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P2i_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P2o_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P3i_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P3o_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P4i_Fz = SubCCSView2DArray(Nfo, Nmz)
        self.P4o_Fz = SubCCSView2DArray(Nfo, Nmz)

        self.Z_P1o = SubCCSView2DArray(Nmz, Nfo)
        self.Z_P2o = SubCCSView2DArray(Nmz, Nfo)
        self.Z_P3o = SubCCSView2DArray(Nmz, Nfo)
        self.Z_P4o = SubCCSView2DArray(Nmz, Nfo)

        self.sig_conn_ptrs.P1i_Fz = <PyObject***>self.P1i_Fz.views
        self.sig_conn_ptrs.P1o_Fz = <PyObject***>self.P1o_Fz.views
        self.sig_conn_ptrs.P2i_Fz = <PyObject***>self.P2i_Fz.views
        self.sig_conn_ptrs.P2o_Fz = <PyObject***>self.P2o_Fz.views
        self.sig_conn_ptrs.P3i_Fz = <PyObject***>self.P3i_Fz.views
        self.sig_conn_ptrs.P3o_Fz = <PyObject***>self.P3o_Fz.views
        self.sig_conn_ptrs.P4i_Fz = <PyObject***>self.P4i_Fz.views
        self.sig_conn_ptrs.P4o_Fz = <PyObject***>self.P4o_Fz.views
        self.sig_conn_ptrs.Z_P1o = <PyObject***>self.Z_P1o.views
        self.sig_conn_ptrs.Z_P2o = <PyObject***>self.Z_P2o.views
        self.sig_conn_ptrs.Z_P3o = <PyObject***>self.Z_P3o.views
        self.sig_conn_ptrs.Z_P4o = <PyObject***>self.Z_P4o.views

        self.P1i_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P1o_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P2i_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P2o_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P3i_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P3o_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P4i_Fyaw = SubCCSView2DArray(Nfo, 1)
        self.P4o_Fyaw = SubCCSView2DArray(Nfo, 1)

        self.yaw_P1o = SubCCSView2DArray(1, Nfo)
        self.yaw_P2o = SubCCSView2DArray(1, Nfo)
        self.yaw_P3o = SubCCSView2DArray(1, Nfo)
        self.yaw_P4o = SubCCSView2DArray(1, Nfo)

        self.sig_conn_ptrs.P1i_Fyaw = <PyObject***>self.P1i_Fyaw.views
        self.sig_conn_ptrs.P1o_Fyaw = <PyObject***>self.P1o_Fyaw.views
        self.sig_conn_ptrs.P2i_Fyaw = <PyObject***>self.P2i_Fyaw.views
        self.sig_conn_ptrs.P2o_Fyaw = <PyObject***>self.P2o_Fyaw.views
        self.sig_conn_ptrs.P3i_Fyaw = <PyObject***>self.P3i_Fyaw.views
        self.sig_conn_ptrs.P3o_Fyaw = <PyObject***>self.P3o_Fyaw.views
        self.sig_conn_ptrs.P4i_Fyaw = <PyObject***>self.P4i_Fyaw.views
        self.sig_conn_ptrs.P4o_Fyaw = <PyObject***>self.P4o_Fyaw.views
        self.sig_conn_ptrs.yaw_P1o = <PyObject***>self.yaw_P1o.views
        self.sig_conn_ptrs.yaw_P2o = <PyObject***>self.yaw_P2o.views
        self.sig_conn_ptrs.yaw_P3o = <PyObject***>self.yaw_P3o.views
        self.sig_conn_ptrs.yaw_P4o = <PyObject***>self.yaw_P4o.views

        self.P1i_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P1o_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P2i_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P2o_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P3i_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P3o_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P4i_Fpitch = SubCCSView2DArray(Nfo, 1)
        self.P4o_Fpitch = SubCCSView2DArray(Nfo, 1)

        self.pitch_P1o = SubCCSView2DArray(1, Nfo)
        self.pitch_P2o = SubCCSView2DArray(1, Nfo)
        self.pitch_P3o = SubCCSView2DArray(1, Nfo)
        self.pitch_P4o = SubCCSView2DArray(1, Nfo)

        self.sig_conn_ptrs.P1i_Fpitch = <PyObject***>self.P1i_Fpitch.views
        self.sig_conn_ptrs.P1o_Fpitch = <PyObject***>self.P1o_Fpitch.views
        self.sig_conn_ptrs.P2i_Fpitch = <PyObject***>self.P2i_Fpitch.views
        self.sig_conn_ptrs.P2o_Fpitch = <PyObject***>self.P2o_Fpitch.views
        self.sig_conn_ptrs.P3i_Fpitch = <PyObject***>self.P3i_Fpitch.views
        self.sig_conn_ptrs.P3o_Fpitch = <PyObject***>self.P3o_Fpitch.views
        self.sig_conn_ptrs.P4i_Fpitch = <PyObject***>self.P4i_Fpitch.views
        self.sig_conn_ptrs.P4o_Fpitch = <PyObject***>self.P4o_Fpitch.views
        self.sig_conn_ptrs.pitch_P1o = <PyObject***>self.pitch_P1o.views
        self.sig_conn_ptrs.pitch_P2o = <PyObject***>self.pitch_P2o.views
        self.sig_conn_ptrs.pitch_P3o = <PyObject***>self.pitch_P3o.views
        self.sig_conn_ptrs.pitch_P4o = <PyObject***>self.pitch_P4o.views


cdef class BeamsplitterValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_9 ptr = (&self.R, &self.T, &self.L, &self.phi, &self.Rcx, &self.Rcy, &self.xbeta, &self.ybeta, &self.alpha)
        cdef tuple params = ("R","T","L","phi","Rcx","Rcy","xbeta","ybeta","alpha")
        self.setup(params, sizeof(ptr), <double**>&ptr)

cdef class BeamsplitterWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, CarrierSignalMatrixSimulation sim, refill):
        cdef FrequencyContainer fcnt

        super().__init__(
            owner,
            sim,
            refill,
            refill,
            BeamsplitterOpticalConnections(owner, sim.carrier),
            BeamsplitterSignalConnections(owner, sim.signal) if sim.signal else None,
            BeamsplitterValues()
        )
        # Store direct type cast for C access
        self.boc = self.carrier.connections
        if sim.signal:
            self.bsc = self.signal.connections
        else:
            self.bsc = None
        self.cvalues = self.values

        self.nr1 = refractive_index(owner.p1)
        self.nr2 = refractive_index(owner.p3)

        if owner.alpha.value == 0.0:
            self.cos_alpha = 1
            self.cos_alpha_2 = 1
        else:
            self.cos_alpha = np.cos(float(owner.alpha.value*DEG2RAD))
            self.cos_alpha_2 = np.cos(np.arcsin(self.nr1 / self.nr2 * np.sin(float(owner.alpha.value*DEG2RAD))))

        # tracing node information
        self.P1i_id = sim.carrier.node_id(owner.p1.i)
        self.P1o_id = sim.carrier.node_id(owner.p1.o)
        self.P2i_id = sim.carrier.node_id(owner.p2.i)
        self.P2o_id = sim.carrier.node_id(owner.p2.o)
        self.P3i_id = sim.carrier.node_id(owner.p3.i)
        self.P3o_id = sim.carrier.node_id(owner.p3.o)
        self.P4i_id = sim.carrier.node_id(owner.p4.i)
        self.P4o_id = sim.carrier.node_id(owner.p4.o)

        if sim.signal:
            self.car_p1o_rhs_idx = sim.carrier.get_node_info(owner.p1.o)['rhs_index']
            self.car_p1i_rhs_idx = sim.carrier.get_node_info(owner.p1.i)['rhs_index']
            self.car_p2o_rhs_idx = sim.carrier.get_node_info(owner.p2.o)['rhs_index']
            self.car_p2i_rhs_idx = sim.carrier.get_node_info(owner.p2.i)['rhs_index']
            self.car_p3o_rhs_idx = sim.carrier.get_node_info(owner.p3.o)['rhs_index']
            self.car_p3i_rhs_idx = sim.carrier.get_node_info(owner.p3.i)['rhs_index']
            self.car_p4o_rhs_idx = sim.carrier.get_node_info(owner.p4.o)['rhs_index']
            self.car_p4i_rhs_idx = sim.carrier.get_node_info(owner.p4.i)['rhs_index']
            self.car_p_num_hom =   sim.carrier.get_node_info(owner.p1.o)['nhoms']

            self.z_signal_enabled = owner.mech.z.full_name in sim.signal.nodes
            if self.z_signal_enabled:
                # Get a reference to the mechanical node frequencies
                fcnt = sim.signal.mechanical_frequencies[owner.mech.z]
                self.z_mech_freqs = fcnt.frequency_info
                self.z_mech_freqs_size = sim.signal.mechanical_frequencies[owner.mech.z].size

            self.yaw_signal_enabled = owner.mech.yaw.full_name in sim.signal.nodes
            if self.yaw_signal_enabled:
                fcnt = sim.signal.mechanical_frequencies[owner.mech.yaw]
                self.yaw_mech_freqs = fcnt.frequency_info
                self.yaw_mech_freqs_size = sim.signal.mechanical_frequencies[owner.mech.yaw].size
                self.K_yaw_sig = make_unscaled_X_scatter_knm_matrix(self.sim.model_data.homs_view)

            self.pitch_signal_enabled = owner.mech.pitch.full_name in sim.signal.nodes
            if self.pitch_signal_enabled:
                fcnt = sim.signal.mechanical_frequencies[owner.mech.pitch]
                self.pitch_mech_freqs = fcnt.frequency_info
                self.pitch_mech_freqs_size = sim.signal.mechanical_frequencies[owner.mech.pitch].size
                self.K_pitch_sig = make_unscaled_Y_scatter_knm_matrix(self.sim.model_data.homs_view)

        self.sym_abcd_elements[:] = [
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
            <cy_expr**> calloc(4, sizeof(cy_expr*)), <cy_expr**> calloc(4, sizeof(cy_expr*)),
        ]
        self.abcd_elements[:] = [
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
        ]

    def __dealloc__(self):
        cdef Py_ssize_t k, i
        for k in range(16):
            if self.sym_abcd_elements[k] == NULL:
                continue

            for i in range(4):
                cy_expr_free(self.sym_abcd_elements[k][i])

            free(self.sym_abcd_elements[k])
            self.sym_abcd_elements[k] = NULL

    cpdef compile_abcd_cy_exprs(self):
        cdef:
            object beamsplitter = self.owner
            list abcd_handles = list(beamsplitter._abcd_matrices.values())

        if self.sim.is_modal:
            # Check for total reflections first
            if self.abcd_p1p2_x is None:
                self.abcd_elements[0] = NULL
                self.abcd_elements[1] = NULL
            else:
                self.abcd_elements[0] = <double*>&self.abcd_p1p2_x[0][0]
                self.abcd_elements[1] = <double*>&self.abcd_p1p2_y[0][0]
            if self.abcd_p2p1_x is None:
                self.abcd_elements[2] = NULL
                self.abcd_elements[3] = NULL
            else:
                self.abcd_elements[2] = <double*>&self.abcd_p2p1_x[0][0]
                self.abcd_elements[3] = <double*>&self.abcd_p2p1_y[0][0]
            if self.abcd_p3p4_x is None:
                self.abcd_elements[4] = NULL
                self.abcd_elements[5] = NULL
            else:
                self.abcd_elements[4] = <double*>&self.abcd_p3p4_x[0][0]
                self.abcd_elements[5] = <double*>&self.abcd_p3p4_y[0][0]
            if self.abcd_p4p3_x is None:
                self.abcd_elements[6] = NULL
                self.abcd_elements[7] = NULL
            else:
                self.abcd_elements[6] = <double*>&self.abcd_p4p3_x[0][0]
                self.abcd_elements[7] = <double*>&self.abcd_p4p3_y[0][0]

            self.abcd_elements[8:16] = [
                <double*>&self.abcd_p1p3_x[0][0], <double*>&self.abcd_p1p3_y[0][0],
                <double*>&self.abcd_p3p1_x[0][0], <double*>&self.abcd_p3p1_y[0][0],
                <double*>&self.abcd_p2p4_x[0][0], <double*>&self.abcd_p2p4_y[0][0],
                <double*>&self.abcd_p4p2_x[0][0], <double*>&self.abcd_p4p2_y[0][0],
            ]

        cdef Py_ssize_t k, i, j
        cdef object[:, ::1] M_sym
        for k in range(16):
            if abcd_handles[k][0] is None:
                continue

            M_sym = abcd_handles[k][0]
            for i in range(2):
                for j in range(2):
                    if isinstance(M_sym[i][j], Symbol):
                        ch_sym = M_sym[i][j].expand_symbols().eval(keep_changing_symbols=True)
                        if isinstance(ch_sym, Symbol):
                            self.sym_abcd_elements[k][2 * i + j] = cy_expr_new()
                            cy_expr_init(
                                self.sym_abcd_elements[k][2 * i + j],
                                ch_sym,
                            )

    cpdef update_parameter_values(self):
        ConnectorWorkspace.update_parameter_values(self)

        cdef Py_ssize_t k, i
        for k in range(16):
            if self.abcd_elements[k] == NULL:
                continue

            for i in range(4):
                if self.sym_abcd_elements[k][i] != NULL:
                    self.abcd_elements[k][i] = cy_expr_eval(
                        self.sym_abcd_elements[k][i]
                    )


cdef inline void beamsplitter_fill_optical_2_optical(
        bs_optical_connections *conn,
        BeamsplitterWorkspace ws,
        frequency_info_t *freq,
        double r,
        double t,
        double phi_0,
        double alpha
    ):
    cdef double phase_shift_scaling = (1 + freq.f / ws.sim.model_data.f0)
    # Phase on reflection is not equal if nr1 != nr2 and AoI != 0
    # so the usual i on transmission phase no longer works.
    cdef double phi_r1, phi_r2, phi_t
    cdef complex_t _r1, _r2, _t0
    if ws.sim.model_data.v2_transmission_phase or ws.nr1 == ws.nr2:
        # old v2 phase on transmission
        # The usual i on transmission and reflections
        # are opposite phase on each side
        phi_r1 = 2 * phi_0 * ws.cos_alpha * phase_shift_scaling
        _r1 = r * cexp(1j * phi_r1)
        _r2 = conj(_r1)
        _t0 = 1j * t
    else:
        # Uses N=-1, Eq.2.25 in Living Rev Relativ (2016) 19:3 DOI 10.1007/s41114-016-0002-8
        # bs transmission phase depends on the reflectivity, refractive indices,
        # and angle of incidence
        phi_r1 = 2 * phi_0 * ws.nr1 * ws.cos_alpha * phase_shift_scaling
        phi_r2 = -2 * phi_0 * ws.nr2 * ws.cos_alpha_2 * phase_shift_scaling
        phi_t = PI/2 + 0.5 * (phi_r1 + phi_r2)
        _r1 = r * cexp(1j * phi_r1)
        _r2 = r * cexp(1j * phi_r2)
        _t0 = t * cexp(1j * phi_t)

    # reflections
    if conn.P1i_P2o[freq.index]:
        (<SubCCSView>conn.P1i_P2o[freq.index]).fill_negative_za_zm_2(_r1, ws.K12.mtx.ptr, ws.K12.mtx.stride1, ws.K12.mtx.stride2)
    if conn.P2i_P1o[freq.index]:
        (<SubCCSView>conn.P2i_P1o[freq.index]).fill_negative_za_zm_2(_r1, ws.K21.mtx.ptr, ws.K21.mtx.stride1, ws.K21.mtx.stride2)
    if conn.P3i_P4o[freq.index]:
        (<SubCCSView>conn.P3i_P4o[freq.index]).fill_negative_za_zm_2(_r2, ws.K34.mtx.ptr, ws.K34.mtx.stride1, ws.K34.mtx.stride2)
    if conn.P4i_P3o[freq.index]:
        (<SubCCSView>conn.P4i_P3o[freq.index]).fill_negative_za_zm_2(_r2, ws.K43.mtx.ptr, ws.K43.mtx.stride1, ws.K43.mtx.stride2)
    # transmissions
    if conn.P1i_P3o[freq.index]:
        (<SubCCSView>conn.P1i_P3o[freq.index]).fill_negative_za_zm_2(_t0, ws.K13.mtx.ptr, ws.K13.mtx.stride1, ws.K13.mtx.stride2)
    if conn.P3i_P1o[freq.index]:
        (<SubCCSView>conn.P3i_P1o[freq.index]).fill_negative_za_zm_2(_t0, ws.K31.mtx.ptr, ws.K31.mtx.stride1, ws.K31.mtx.stride2)
    if conn.P2i_P4o[freq.index]:
        (<SubCCSView>conn.P2i_P4o[freq.index]).fill_negative_za_zm_2(_t0, ws.K24.mtx.ptr, ws.K24.mtx.stride1, ws.K24.mtx.stride2)
    if conn.P4i_P2o[freq.index]:
        (<SubCCSView>conn.P4i_P2o[freq.index]).fill_negative_za_zm_2(_t0, ws.K42.mtx.ptr, ws.K42.mtx.stride1, ws.K42.mtx.stride2)


beamsplitter_carrier_fill = FillFuncWrapper.make_from_ptr(c_beamsplitter_carrier_fill)
cdef object c_beamsplitter_carrier_fill(ConnectorWorkspace cws):
    r"""
    Fills the sub-matrix of the interferometer matrix held by `sim`, corresponding
    to the `beamsplitter` component.

    A beam splitter is similar to a mirror except for the extra parameter :math:`\alpha`,
    which indicates the angle of incidence of the incoming beams, and that it has four
    ports with two couplings each. This is shown in :numref:`fig_bs_couplings`. See
    :func:`mirror_fill` for details on the common surface arguments.

    .. _fig_bs_couplings:
    .. figure:: ../images/beamsplitter.*
        :align: center

        Field couplings at a beam splitter with a representation of the reference plane and
        the angle of incidence.

    Displacement of the beam splitter is assumed to be perpendicular to its optical surface,
    therefore the angle of incidence affects the phase change of the reflected light. This phase
    shift is given by,

    .. math::
        \varphi = 2\phi\frac{\omega}{\omega_0} \cos{\left(\alpha\right)},

    where :math:`omega` is the angular frequency of the reflected light.

    From this (and the details given in :func:`mirror_fill`), for each frequency light field :math:`f`
    in the interferometer, the following quantities are computed in general (including higher-order
    spatial modes) for the field couplings,

    .. math::
        \begin{array}{l}
            \mathrm{bs}_{12} = r K_{12} \exp{\left(i 2\phi \left(1 + \frac{f}{f_0}\right) \cos{\alpha} \right)},\\
            \mathrm{bs}_{21} = r K_{21} \exp{\left(i 2\phi \left(1 + \frac{f}{f_0}\right) \cos{\alpha} \right)},\\
            \mathrm{bs}_{13} = it K_{13},\\
            \mathrm{bs}_{31} = it K_{31},\\
            \mathrm{bs}_{34} = r K_{34} \exp{\left(-i 2\phi \left(1 + \frac{f}{f_0}\right) \cos{\alpha} \right)},\\
            \mathrm{bs}_{43} = r K_{43} \exp{\left(-i 2\phi \left(1 + \frac{f}{f_0}\right) \cos{\alpha} \right)},\\
            \mathrm{bs}_{24} = it K_{24},\\
            \mathrm{bs}_{42} = it K_{42},
        \end{array}

    where :math:`K_{\mathrm{ij}}` are the scattering matrices for each direction (see
    :ref:`scatter_matrices`). Here, each :math:`\mathrm{bs}_{\mathrm{ij}}` term now represents a
    vector of the couplings of all higher-order spatial mode fields present.

    Parameters
    ----------
    beamsplitter : :class:`.Beamsplitter`
        The beamsplitter object to fill.

    sim : :class:`.CarrierSignalMatrixSimulation`
        A handle to the simulation.

    values : dict
        Dictionary of evaluated model parameters.

    cos_alpha : float
        Cosine of the angle of incidence at the front surface.

    cos_alpha_2 : float
        Cosine of the angle of incidence at the back surface.

    lambda0 : float
        Wavelength of the laser beam light for the model.
    """
    cdef:
        BeamsplitterWorkspace ws = <BeamsplitterWorkspace>cws
        double f0 = ws.sim.model_data.f0
        double k0 = ws.sim.model_data.k0

        double r = sqrt(ws.cvalues.R)
        double t = sqrt(ws.cvalues.T)
        double phi = DEG2RAD * ws.cvalues.phi
        double alpha = DEG2RAD * ws.cvalues.alpha
        double phase_shift_scaling

        Py_ssize_t _i, size
        frequency_info_t *frequencies
        bs_optical_connections *conn = &ws.boc.opt_conn_ptrs

    size = ws.sim.carrier.optical_frequencies.size
    frequencies = ws.sim.carrier.optical_frequencies.frequency_info

    for i in range(size):
        beamsplitter_fill_optical_2_optical(
            conn, ws, &frequencies[i], r, t, phi, alpha
        )


beamsplitter_signal_fill = FillFuncWrapper.make_from_ptr(c_beamsplitter_signal_fill)
cdef object c_beamsplitter_signal_fill(ConnectorWorkspace cws):
    cdef:
        BeamsplitterWorkspace ws = <BeamsplitterWorkspace>cws
        double r = sqrt(ws.cvalues.R)
        double t = sqrt(ws.cvalues.T)
        double phi = DEG2RAD * ws.cvalues.phi
        double alpha = DEG2RAD * ws.cvalues.alpha
        Py_ssize_t _i, size

        frequency_info_t *frequencies
        bs_optical_connections *car_conn = &ws.boc.opt_conn_ptrs
        bs_optical_connections *conn = &ws.bsc.opt_conn_ptrs
        bs_signal_connections *sconn = &ws.bsc.sig_conn_ptrs

    ws.z_to_field1 = 1j * ws.cos_alpha * ws.sim.model_data.k0 * ws.sim.model_data.x_scale
    ws.z_to_field2 = 1j * ws.cos_alpha_2 * ws.sim.model_data.k0 * ws.sim.model_data.x_scale
    ws.field1_to_F = ws.cos_alpha / (C_LIGHT * ws.sim.model_data.x_scale)
    ws.field2_to_F = ws.cos_alpha_2 / (C_LIGHT * ws.sim.model_data.x_scale)
    size = ws.sim.signal.optical_frequencies.size
    frequencies = ws.sim.signal.optical_frequencies.frequency_info

    for i in range(size):
        beamsplitter_fill_optical_2_optical(
            conn, ws, &frequencies[i], r, t, phi, alpha
        )

    if ws.z_signal_enabled:
        for i in range(size):
            freq = &(frequencies[i])
            if ws.z_mech_freqs_size == 1:
                single_z_mechanical_frequency_signal_calc(ws, sconn, car_conn, freq, phi, 0, freq.audio_carrier_index)
            else:
                multiple_z_mechanical_freq_signal_calc(ws, sconn, car_conn, freq, phi)

    if ws.yaw_signal_enabled:
        for i in range(size):
            freq = &(frequencies[i])
            if ws.yaw_mech_freqs_size == 1:
                single_yaw_mechanical_frequency_signal_calc(ws, sconn, car_conn, freq, phi, 0, freq.audio_carrier_index)
            else:
                raise NotImplementedError()

    if ws.pitch_signal_enabled:
        for i in range(size):
            freq = &(frequencies[i])
            if ws.pitch_mech_freqs_size == 1:
                single_pitch_mechanical_frequency_signal_calc(ws, sconn, car_conn, freq, phi, 0, freq.audio_carrier_index)
            else:
                raise NotImplementedError()


cdef void get_carrier_vectors(BeamsplitterWorkspace ws, int carrier_index,
        DenseZVector *c_p1_i, DenseZVector *c_p2_i,
        DenseZVector *c_p3_i, DenseZVector *c_p4_i,
        DenseZVector *c_p1_o, DenseZVector *c_p2_o,
        DenseZVector *c_p3_o, DenseZVector *c_p4_o
    ):
    assert(c_p1_i)
    assert(c_p2_i)
    assert(c_p3_i)
    assert(c_p4_i)
    assert(c_p1_o)
    assert(c_p2_o)
    assert(c_p3_o)
    assert(c_p4_o)

    c_p1_i.size = c_p1_o.size = c_p2_i.size = c_p2_o.size = ws.car_p_num_hom
    c_p1_i.stride = c_p1_o.stride = c_p2_i.stride = c_p2_o.stride = 1
    c_p3_i.size = c_p3_o.size = c_p4_i.size = c_p4_o.size = ws.car_p_num_hom
    c_p3_i.stride = c_p3_o.stride = c_p4_i.stride = c_p4_o.stride = 1
    # Get incoming/outgoing carrier field amplitudes
    c_p1_i.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1i_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p2_i.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2i_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p3_i.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p3i_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p4_i.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p4i_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p1_o.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p1o_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p2_o.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p2o_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p3_o.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p3o_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]
    c_p4_o.ptr = &ws.sim.carrier.out_view[ws.sim.carrier.field_fast_2(
        ws.car_p4o_rhs_idx, ws.car_p_num_hom, carrier_index, 0
    )]


cdef void multiple_z_mechanical_freq_signal_calc (
        BeamsplitterWorkspace ws,
        bs_signal_connections *conn,
        bs_optical_connections *car_conn,
        frequency_info_t *freq,
        double phi
    ):
    """Computes the opto-mechanics for a mirror with multiple optical and mechanical frequencies.
    """

    cdef:
        Py_ssize_t i, j
        frequency_info_t *ofrequencies = ws.sim.carrier.optical_frequencies.frequency_info
        Py_ssize_t osize = ws.sim.carrier.optical_frequencies.size
        double fs, fc, fm

    for i in range(osize): # Loop over optical DC
        for j in range(ws.z_mech_freqs_size): # Loop over mechanical frequencies
            fs = freq.f
            fc = ofrequencies[i].f
            fm = ws.z_mech_freqs[j].f

            if (fc-fs == fm) or (fs-fc == fm):
                single_z_mechanical_frequency_signal_calc(
                    ws, conn, car_conn, freq, phi, j, i
                )


cdef void single_z_mechanical_frequency_signal_calc (
        BeamsplitterWorkspace ws,
        bs_signal_connections *conn,
        bs_optical_connections *car_conn,
        frequency_info_t *freq,
        double phi,
        Py_ssize_t z_freq_idx,
        Py_ssize_t carrier_index
    ):
    cdef:
        complex_t _tuning, _ctuning
        DenseZVector c_p1_i, c_p2_i, c_p1_o, c_p2_o
        DenseZVector c_p3_i, c_p4_i, c_p3_o, c_p4_o

    get_carrier_vectors(ws, carrier_index,
        &c_p1_i, &c_p2_i, &c_p3_i, &c_p4_i,
        &c_p1_o, &c_p2_o, &c_p3_o, &c_p4_o,
    )

    # -------------------------------------------------
    # Optical to mechanical connections
    # -------------------------------------------------
    # - Longitudinal
    # -------------------------------------------------
    # These fill a nHOMx1 matrix to compute RP force
    if conn.P1i_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P1i_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            -ws.field1_to_F, c_p1_i.ptr, 1, 1
        )
    if conn.P1o_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P1o_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            -ws.field1_to_F, c_p1_o.ptr, 1, 1
        )
    if conn.P2i_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P2i_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            -ws.field1_to_F, c_p2_i.ptr, 1, 1
        )
    if conn.P2o_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P2o_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            -ws.field1_to_F, c_p2_o.ptr, 1, 1
        )

    # Minus sign as we force the mirror in the opposite
    # direction from the other side
    if conn.P3i_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P3i_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            ws.field2_to_F, c_p3_i.ptr, 1, 1
        )
    if conn.P3o_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P3o_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            ws.field2_to_F, c_p3_o.ptr, 1, 1
        )
    if conn.P4i_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P4i_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            ws.field2_to_F, c_p4_i.ptr, 1, 1
        )
    if conn.P4o_Fz[freq.index][z_freq_idx]:
        (<SubCCSView>conn.P4o_Fz[freq.index][z_freq_idx]).fill_negative_za_zmc (
            ws.field2_to_F, c_p4_o.ptr, 1, 1
        )

    # -----------------------------------------------------------------
    # Mechanical to optical connections
    # -----------------------------------------------------------------
    # - Longitudinal
    # -----------------------------------------------------------------
    # As the output has a mixture of both refl and transmitted we only
    # modulate the incoming and refl'd field so we have to propagate
    # the input

    # As we are using the propagaged carrier, it already has the various phase
    # static phase+amplitude factors, HOM scattering, etc. included, which is
    # useful as that means we don't duplicate the calculations here. However,
    # the phase accumulated is slightly different, as frequency shift happens
    # at the mirror so it picks up a slightly different detuning phase coming
    # back from the mirror, here we correct that
    phase_shift = phi * freq.f / ws.sim.model_data.f0

    # -----------------------------------------------------------------
    # Signal generation z->p1.o
    _tuning = cexp(1.0j * phase_shift)
    if conn.Z_P1o[z_freq_idx][freq.index]:
        # fill_prop_za as off-diagonal -1 is already included in the carrier connection
        (<SubCCSView>conn.Z_P1o[z_freq_idx][freq.index]).fill_prop_za (
            (<SubCCSView>car_conn.P2i_P1o[carrier_index]), 0, ws.z_to_field1 * _tuning, False
        )
    if conn.Z_P2o[z_freq_idx][freq.index]:
        # fill_prop_za as off-diagonal -1 is already included in the carrier connection
        (<SubCCSView>conn.Z_P2o[z_freq_idx][freq.index]).fill_prop_za (
            (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0, ws.z_to_field1 * _tuning, False
        )
    # -----------------------------------------------------------------
    # Signal generation z->p2.o
    # extra 180 phase here as we're doing the opposite
    # modulation when looked at from the other side of the mirror
    if conn.Z_P3o[z_freq_idx][freq.index]:
        _ctuning = conj(_tuning)
        # fill_prop_za as off-diagonal -1 is already included in the carrier connection
        (<SubCCSView>conn.Z_P3o[z_freq_idx][freq.index]).fill_prop_za (
            (<SubCCSView>car_conn.P4i_P3o[carrier_index]), 0, -ws.z_to_field2 * _ctuning, False
        )
    if conn.Z_P4o[z_freq_idx][freq.index]:
        _ctuning = conj(_tuning)
        # fill_prop_za as off-diagonal -1 is already included in the carrier connection
        (<SubCCSView>conn.Z_P4o[z_freq_idx][freq.index]).fill_prop_za (
            (<SubCCSView>car_conn.P3i_P4o[carrier_index]), 0, -ws.z_to_field2 * _ctuning, False
        )


cdef void single_yaw_mechanical_frequency_signal_calc (
        BeamsplitterWorkspace ws,
        bs_signal_connections *conn,
        bs_optical_connections *car_conn,
        frequency_info_t *freq, # audio sideband
        double phi,
        Py_ssize_t yaw_freq_idx,
        Py_ssize_t carrier_index
    ):
    cdef:
        complex_t tuning, ctuning
        double wx1i, wx2i, wx3i, wx4i
        double wx1o = 0.0
        double wx2o = 0.0
        double wx3o = 0.0
        double wx4o = 0.0
        NodeBeamParam *q_P1o
        NodeBeamParam *q_P1i
        NodeBeamParam *q_P2o
        NodeBeamParam *q_P2i
        NodeBeamParam *q_P3o
        NodeBeamParam *q_P3i
        NodeBeamParam *q_P4o
        NodeBeamParam *q_P4i
        complex_t a1_2_o_factor, a2_2_o_factor
        complex_t phase_shift = cexp(1j*phi * freq.f / ws.sim.model_data.f0)
        DenseZVector c_p1_i, c_p2_i, c_p1_o, c_p2_o
        DenseZVector c_p3_i, c_p4_i, c_p3_o, c_p4_o

    get_carrier_vectors(ws, carrier_index,
        &c_p1_i, &c_p2_i, &c_p3_i, &c_p4_i,
        &c_p1_o, &c_p2_o, &c_p3_o, &c_p4_o,
    )

    # We use an unscaled Knm matrix, so we need to apply the waist size and gouy phase
    # as we always reverse the gouy phase anyway, we just don't bother adding it here
    # k0 scaling with nr is done in code jsut below, along with spot size
    # TODO ddb - these matrix multiplications would be more efficient with a sparse matrix
    # format, CSR maybe? As dense product scales badly with maxtem
    if ws.sim.is_modal: # ignore filling this if doing plane-wave
        a1_2_o_factor = 1j * ws.cos_alpha * ws.sim.model_data.k0 * ws.sim.model_data.x_scale * (1 + freq.f_car[0]/ws.sim.model_data.f0)
        a2_2_o_factor = 1j * ws.cos_alpha_2 * ws.sim.model_data.k0 * ws.sim.model_data.x_scale * (1 + freq.f_car[0]/ws.sim.model_data.f0)
        q_P1o = &ws.sim.trace[ws.P1o_id]
        q_P2o = &ws.sim.trace[ws.P2o_id]
        q_P3o = &ws.sim.trace[ws.P3o_id]
        q_P4o = &ws.sim.trace[ws.P4o_id]
        q_P1i = &ws.sim.trace[ws.P1i_id]
        q_P2i = &ws.sim.trace[ws.P2i_id]
        q_P3i = &ws.sim.trace[ws.P3i_id]
        q_P4i = &ws.sim.trace[ws.P4i_id]

        if conn.yaw_P1o[yaw_freq_idx][freq.index]:
            wx1o = bp_beamsize(&q_P1o.qx)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.yaw_P1o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P2i_P1o[carrier_index]), 0,
                -2/2*ws.nr1 * wx1o * a1_2_o_factor * phase_shift,
                ws.K_yaw_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.yaw_P1o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P3i_P1o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wx1o * a1_2_o_factor,
                ws.K_yaw_sig.data_view, True
            )

        if conn.yaw_P2o[yaw_freq_idx][freq.index]:
            wx2o = bp_beamsize(&q_P2o.qx)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.yaw_P2o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0,
                -2/2*ws.nr1 * wx2o * a1_2_o_factor * phase_shift,
                ws.K_yaw_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.yaw_P2o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wx2o * a1_2_o_factor,
                ws.K_yaw_sig.data_view, True
            )

        if conn.yaw_P3o[yaw_freq_idx][freq.index]:
            wx3o = bp_beamsize(&q_P3o.qx)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.yaw_P3o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P4i_P3o[carrier_index]), 0,
                -2/2*ws.nr2 * wx3o * a2_2_o_factor * conj(phase_shift),
                ws.K_yaw_sig.data_view, False
            )
            # Transmission coupling
            (<SubCCSView>conn.yaw_P3o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P1i_P3o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wx3o * a2_2_o_factor,
                ws.K_yaw_sig.data_view, True
            )

        if conn.yaw_P4o[yaw_freq_idx][freq.index]:
            wx4o = bp_beamsize(&q_P4o.qx)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.yaw_P4o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P3i_P4o[carrier_index]), 0,
                -2/2*ws.nr2 * wx4o * a2_2_o_factor * conj(phase_shift),
                ws.K_yaw_sig.data_view, False
            )
            # Transmission coupling
            (<SubCCSView>conn.yaw_P4o[yaw_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P2i_P4o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wx4o * a2_2_o_factor,
                ws.K_yaw_sig.data_view, True
            )

        # -------------------------------------------------
        # Optical to mechanical connections
        # -------------------------------------------------
        # These fill a nHOMx1 matrix to compute RP force
        # There is a minus sign difference between side 1 and 2 here, because
        # of the coordinate system change
        if conn.P1i_Fyaw[freq.index][yaw_freq_idx]:
            wx1i = bp_beamsize(&q_P1i.qx)
            (<SubCCSView>conn.P1i_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                ws.nr1 * wx1i * ws.field1_to_F, &ws.K_yaw_sig.mtx, &c_p1_i
            )
        # differing minus signs here because of the x coordinate flip compared to mechanical node x
        if conn.P1o_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P1o_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wx1o * ws.field1_to_F, &ws.K_yaw_sig.mtx, &c_p1_o
            )

        if conn.P2i_Fyaw[freq.index][yaw_freq_idx]:
            wx2i = bp_beamsize(&q_P2i.qx)
            (<SubCCSView>conn.P2i_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                ws.nr1 * wx2i * ws.field1_to_F, &ws.K_yaw_sig.mtx, &c_p2_i
            )
        if conn.P2o_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P2o_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wx2o * ws.field1_to_F, &ws.K_yaw_sig.mtx, &c_p2_o
            )

        if conn.P3i_Fyaw[freq.index][yaw_freq_idx]:
            wx3i = bp_beamsize(&q_P3i.qx)
            (<SubCCSView>conn.P3i_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wx3i * ws.field2_to_F, &ws.K_yaw_sig.mtx,  &c_p3_i
            )
        if conn.P3o_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P3o_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                -ws.nr2 * wx3o * ws.field2_to_F, &ws.K_yaw_sig.mtx, &c_p3_o
            )

        if conn.P4i_Fyaw[freq.index][yaw_freq_idx]:
            wx4i = bp_beamsize(&q_P4i.qx)
            (<SubCCSView>conn.P4i_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wx4i * ws.field2_to_F, &ws.K_yaw_sig.mtx,  &c_p4_i
            )
        if conn.P4o_Fyaw[freq.index][yaw_freq_idx]:
            (<SubCCSView>conn.P4o_Fyaw[freq.index][yaw_freq_idx]).fill_negative_za_zmvc (
                -ws.nr2 * wx4o * ws.field2_to_F, &ws.K_yaw_sig.mtx, &c_p4_o
            )


cdef void single_pitch_mechanical_frequency_signal_calc (
        BeamsplitterWorkspace ws,
        bs_signal_connections *conn,
        bs_optical_connections *car_conn,
        frequency_info_t *freq, # audio sideband
        double phi,
        Py_ssize_t pitch_freq_idx,
        Py_ssize_t carrier_index
    ):
    cdef:
        complex_t tuning, ctuning
        double wy1i, wy2i, wy3i, wy4i
        double wy1o = 0.0
        double wy2o = 0.0
        double wy3o = 0.0
        double wy4o = 0.0
        NodeBeamParam *q_P1o
        NodeBeamParam *q_P1i
        NodeBeamParam *q_P2o
        NodeBeamParam *q_P2i
        NodeBeamParam *q_P3o
        NodeBeamParam *q_P3i
        NodeBeamParam *q_P4o
        NodeBeamParam *q_P4i
        complex_t a1_2_o_factor, a2_2_o_factor
        complex_t phase_shift = cexp(1j*phi * freq.f / ws.sim.model_data.f0)
        DenseZVector c_p1_i, c_p2_i, c_p1_o, c_p2_o
        DenseZVector c_p3_i, c_p4_i, c_p3_o, c_p4_o

    get_carrier_vectors(ws, carrier_index,
        &c_p1_i, &c_p2_i, &c_p3_i, &c_p4_i,
        &c_p1_o, &c_p2_o, &c_p3_o, &c_p4_o,
    )

    if ws.sim.is_modal: # ignore filling this if doing plane-wave
        a1_2_o_factor = 1j * ws.cos_alpha * ws.sim.model_data.k0 * ws.sim.model_data.x_scale * (1 + freq.f_car[0]/ws.sim.model_data.f0)
        a2_2_o_factor = 1j * ws.cos_alpha_2 * ws.sim.model_data.k0 * ws.sim.model_data.x_scale * (1 + freq.f_car[0]/ws.sim.model_data.f0)
        q_P1o = &ws.sim.trace[ws.P1o_id]
        q_P2o = &ws.sim.trace[ws.P2o_id]
        q_P3o = &ws.sim.trace[ws.P3o_id]
        q_P4o = &ws.sim.trace[ws.P4o_id]
        q_P1i = &ws.sim.trace[ws.P1i_id]
        q_P2i = &ws.sim.trace[ws.P2i_id]
        q_P3i = &ws.sim.trace[ws.P3i_id]
        q_P4i = &ws.sim.trace[ws.P4i_id]

        if conn.pitch_P1o[pitch_freq_idx][freq.index]:
            wy1o = bp_beamsize(&q_P1o.qy)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.pitch_P1o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P2i_P1o[carrier_index]), 0,
                -2/2*ws.nr1 * wy1o * a1_2_o_factor * phase_shift,
                ws.K_pitch_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.pitch_P1o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P3i_P1o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wy1o * a1_2_o_factor,
                ws.K_pitch_sig.data_view, True
            )

        if conn.pitch_P2o[pitch_freq_idx][freq.index]:
            wy2o = bp_beamsize(&q_P2o.qy)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.pitch_P2o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0,
                -2/2*ws.nr1 * wy2o * a1_2_o_factor * phase_shift,
                ws.K_pitch_sig.data_view, False
            )
            # Transmission
            (<SubCCSView>conn.pitch_P2o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P1i_P2o[carrier_index]), 0,
                -0.5 * (ws.nr1 - ws.nr2) * wy2o * a1_2_o_factor,
                ws.K_pitch_sig.data_view, True
            )

        if conn.pitch_P3o[pitch_freq_idx][freq.index]:
            wy3o = bp_beamsize(&q_P3o.qy)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.pitch_P3o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P4i_P3o[carrier_index]), 0,
                2/2*ws.nr2 * wy3o * a2_2_o_factor * conj(phase_shift),
                ws.K_pitch_sig.data_view, False
            )
            # Transmission coupling
            (<SubCCSView>conn.pitch_P3o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P1i_P3o[carrier_index]), 0,
                0.5 * (ws.nr1 - ws.nr2) * wy3o * a2_2_o_factor,
                ws.K_pitch_sig.data_view, True
            )

        if conn.pitch_P4o[pitch_freq_idx][freq.index]:
            wy4o = bp_beamsize(&q_P4o.qy)
            # fill_prop_za as off-diagonal -1 is already included in the carrier connection
            (<SubCCSView>conn.pitch_P4o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                # factor of 2 because misalignment is 2 * x/ybeta, but 0.5 factor from upper/lower SB gain
                (<SubCCSView>car_conn.P3i_P4o[carrier_index]), 0,
                2/2*ws.nr2 * wy4o * a2_2_o_factor * conj(phase_shift),
                ws.K_pitch_sig.data_view, False
            )
            # Transmission coupling
            (<SubCCSView>conn.pitch_P4o[pitch_freq_idx][freq.index]).fill_prop_za_zm (
                (<SubCCSView>car_conn.P2i_P4o[carrier_index]), 0,
                0.5 * (ws.nr1 - ws.nr2) * wy4o * a2_2_o_factor,
                ws.K_pitch_sig.data_view, True
            )

        # -------------------------------------------------
        # Optical to mechanical connections
        # -------------------------------------------------
        # These fill a nHOMx1 matrix to compute RP force
        # There is a minus sign difference between side 1 and 2 here, because
        # of the coordinate system change
        if conn.P1i_Fpitch[freq.index][pitch_freq_idx]:
            wy1i = bp_beamsize(&q_P1i.qy)
            (<SubCCSView>conn.P1i_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wy1i * ws.field1_to_F, &ws.K_pitch_sig.mtx, &c_p1_i
            )
        # differing minus signs here because of the x coordinate flip compared to mechanical node x
        if conn.P1o_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P1o_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wy1o * ws.field1_to_F, &ws.K_pitch_sig.mtx, &c_p1_o
            )

        if conn.P2i_Fpitch[freq.index][pitch_freq_idx]:
            wy2i = bp_beamsize(&q_P2i.qy)
            (<SubCCSView>conn.P2i_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wy2i * ws.field1_to_F, &ws.K_pitch_sig.mtx, &c_p2_i
            )
        if conn.P2o_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P2o_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                -ws.nr1 * wy2o * ws.field1_to_F, &ws.K_pitch_sig.mtx, &c_p2_o
            )

        if conn.P3i_Fpitch[freq.index][pitch_freq_idx]:
            wy3i = bp_beamsize(&q_P3i.qy)
            (<SubCCSView>conn.P3i_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wy3i * ws.field2_to_F, &ws.K_pitch_sig.mtx,  &c_p3_i
            )
        if conn.P3o_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P3o_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wy3o * ws.field2_to_F, &ws.K_pitch_sig.mtx, &c_p3_o
            )

        if conn.P4i_Fpitch[freq.index][pitch_freq_idx]:
            wy4i = bp_beamsize(&q_P4i.qy)
            (<SubCCSView>conn.P4i_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wy4i * ws.field2_to_F, &ws.K_pitch_sig.mtx,  &c_p4_i
            )
        if conn.P4o_Fpitch[freq.index][pitch_freq_idx]:
            (<SubCCSView>conn.P4o_Fpitch[freq.index][pitch_freq_idx]).fill_negative_za_zmvc (
                ws.nr2 * wy4o * ws.field2_to_F, &ws.K_pitch_sig.mtx, &c_p4_o
            )


beamsplitter_fill_qnoise = FillFuncWrapper.make_from_ptr(c_beamsplitter_fill_qnoise)
cdef object c_beamsplitter_fill_qnoise(ConnectorWorkspace cws):
    r"""
    Fills the quantum noise input matrix elements corresponding to this `beamsplitter`.
    """
    cdef:
        BeamsplitterWorkspace ws = <BeamsplitterWorkspace> cws
        PyObject ***noises = ws.output_noise.ptrs
        frequency_info_t *freq

        Py_ssize_t i, j

        double qn_internal_loss
        complex_t factor

    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        factor = 0.5 * (1 + freq.f_car[0] / ws.sim.model_data.f0)
        qn_internal_loss = ws.cvalues.L
        if not ws.sim.is_modal:
            (<SubCCSView>noises[0][freq.index]).fill_za(factor * qn_internal_loss)
            (<SubCCSView>noises[1][freq.index]).fill_za(factor * qn_internal_loss)
            (<SubCCSView>noises[2][freq.index]).fill_za(factor * qn_internal_loss)
            (<SubCCSView>noises[3][freq.index]).fill_za(factor * qn_internal_loss)
        else:
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += qn_internal_loss
                ws.total_losses[j] += ws.cvalues.R * ws.oconn_info[1].loss[j]
                ws.total_losses[j] += ws.cvalues.T * ws.oconn_info[6].loss[j]
            (<SubCCSView>noises[0][freq.index]).fill_za_dd(factor, ws.total_losses)
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += qn_internal_loss
                ws.total_losses[j] += ws.cvalues.R * ws.oconn_info[0].loss[j]
                ws.total_losses[j] += ws.cvalues.T * ws.oconn_info[7].loss[j]
            (<SubCCSView>noises[1][freq.index]).fill_za_dd(factor, ws.total_losses)
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += qn_internal_loss
                ws.total_losses[j] += ws.cvalues.R * ws.oconn_info[3].loss[j]
                ws.total_losses[j] += ws.cvalues.T * ws.oconn_info[4].loss[j]
            (<SubCCSView>noises[2][freq.index]).fill_za_dd(factor, ws.total_losses)
            ws.total_losses[:] = 0
            for j in range(ws.sim.signal.nhoms):
                ws.total_losses[j] += qn_internal_loss
                ws.total_losses[j] += ws.cvalues.R * ws.oconn_info[2].loss[j]
                ws.total_losses[j] += ws.cvalues.T * ws.oconn_info[5].loss[j]
            (<SubCCSView>noises[3][freq.index]).fill_za_dd(factor, ws.total_losses)



from finesse.cymath cimport complex_t
from finesse.cyexpr cimport (
    cy_expr,
    cy_expr_new,
    cy_expr_init,
    cy_expr_free,
    cy_expr_eval,
)
from finesse.element cimport ElementWorkspace
from finesse.tracing.ctracer cimport TraceTree
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation

cdef enum cavity_plane:
    X,
    Y


cdef class CavityWorkspace(ElementWorkspace):
    cdef:
        # Round-trip ABCDs in both planes
        double[:, ::1] abcd_x
        double[:, ::1] abcd_y

        # The internal cavity tree (grabbed from the
        # simulations' changing forest)
        TraceTree tree

        # Flag for whether the internal cavity tree is
        # in the changing forest of the simulation
        bint is_changing
        bint is_stable # is 0 < {gx, gy} < 1
        bint is_stable_x # is 0 < gx < 1
        bint is_stable_y # is 0 < gy < 1

        # Store both symbolic and numeric expressions for
        # the relevant cavity parameters - the symbolic
        # fields consist of only the changing symbols if any
        cy_expr* sym_length
        double length

        cy_expr* sym_loss
        double loss

        cy_expr* sym_finesse
        double finesse

        cy_expr* sym_fsr
        double fsr

        cy_expr* sym_fwhm
        double fwhm

        cy_expr* sym_tau
        double tau

        cy_expr* sym_pole
        double pole

        # These parameters are dependent upon the round-trip ABCD
        # so are computed differently (via the cavity trace tree)
        # for efficiency

        # round-trip Gouy phase
        double rt_gouy_x
        double rt_gouy_y

        # mode separation frequency
        double Df_x
        double Df_y

        # eigenmode
        complex_t qx
        complex_t qy

        # resolution
        double Sx
        double Sy

        # stability (g-factor)
        double gx
        double gy


    cdef void __update_geometric(self)

    cdef void update(self)
    cdef bint update_eigenmode(self, cavity_plane plane)
    cdef void update_stability(self, cavity_plane plane)
    cdef void update_rt_gouy(self, cavity_plane plane)
    cdef void update_Df(self, cavity_plane plane)
    cdef void update_S(self, cavity_plane plane)



from libc.stdlib cimport calloc, free

cimport cython

cimport numpy as np
import numpy as np

from finesse.cymath cimport complex_t
from finesse.cymath.math cimport degrees
from finesse.cymath.gaussbeam cimport (
    bp_beamsize,
    bp_waistsize,
    bp_waistpos,
    bp_rayleigh,
    bp_gouy,
    bp_divergence,
    bp_radius_curvature,
    bp_defocus,
    bp_overlap,
    transform_q,
    overlap,
)
from finesse.detectors.workspace cimport (
    DetectorWorkspace,
    OutputFuncWrapper,
)
from finesse.components.workspace cimport ConnectorWorkspace
from finesse.components.modal.space cimport SpaceWorkspace

from finesse.gaussian import BeamParam


### Beam Property Detector ###


cdef class BPDetectorWorkspace(DetectorWorkspace):
    """Workspace for beam property calculations used by :class:`.BeamPropertyDetector`."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        cdef Py_ssize_t n_id = self.sim.carrier.node_id(owner.node)
        if owner.direction == "x":
            self.q = &self.sim.trace[n_id].qx
        else:
            self.q = &self.sim.trace[n_id].qy

        self.is_changing = self.sim.trace[n_id].is_changing

        self.detecting = owner.detecting
        if self.detecting == BeamProperty.Q:
            self.compute_func = NULL
            self.initial = -1.0
            self.set_output_fn(bp_q_detector_output)
        else:
            if self.detecting == BeamProperty.SIZE:
                self.compute_func = bp_beamsize
            elif self.detecting == BeamProperty.WAISTSIZE:
                self.compute_func = bp_waistsize
            elif self.detecting == BeamProperty.DISTANCE:
                self.compute_func = bp_waistpos
            elif self.detecting == BeamProperty.RAYLEIGH:
                self.compute_func = bp_rayleigh
            elif self.detecting == BeamProperty.GOUY:
                self.compute_func = bp_gouy
            elif self.detecting == BeamProperty.DIVERGENCE:
                self.compute_func = bp_divergence
            elif self.detecting == BeamProperty.ROC:
                self.compute_func = bp_radius_curvature
            elif self.detecting == BeamProperty.DEFOCUS:
                self.compute_func = bp_defocus
            else:
                raise RuntimeError(f"Unrecognised value for BeamProperty: {self.detecting}")

            self.initial = self.compute_func(self.q)
            self.set_output_fn(bp_detector_output)

bp_detector_output = OutputFuncWrapper.make_from_ptr(c_bp_detector_output)
cdef c_bp_detector_output(DetectorWorkspace dws):
    cdef BPDetectorWorkspace ws = <BPDetectorWorkspace> dws

    if not ws.is_changing:
        return ws.initial

    return ws.compute_func(ws.q)

bp_q_detector_output = OutputFuncWrapper.make_from_ptr(c_bp_q_detector_output)
cdef c_bp_q_detector_output(DetectorWorkspace dws):
    cdef BPDetectorWorkspace ws = <BPDetectorWorkspace> dws

    if ws.q_as_bp:
        return BeamParam(ws.q.wavelength, ws.q.nr, ws.q.q)

    return ws.q.q


### Cavity Detector ###


cdef class CPDetectorWorkspace(DetectorWorkspace):
    """Workspace for cavity property calculations, used by :class:`.CavityPropertyDetector`,
    for scalar properties which do not rely on the cavity eigenmode."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        cdef CavityWorkspace cav_ws = self.sim.cavity_workspaces[owner.cavity]
        self.detecting = owner.detecting

        # Set the target ptr
        if self.detecting == CavityProperty.LENGTH:
            self.target = &cav_ws.length
        elif self.detecting == CavityProperty.LOSS:
            self.target = &cav_ws.loss
        elif self.detecting == CavityProperty.FINESSE:
            self.target = &cav_ws.finesse
        elif self.detecting == CavityProperty.FSR:
            self.target = &cav_ws.fsr
        elif self.detecting == CavityProperty.FWHM:
            self.target = &cav_ws.fwhm
        elif self.detecting == CavityProperty.POLE:
            self.target = &cav_ws.pole
        elif self.detecting == CavityProperty.TAU:
            self.target = &cav_ws.tau
        elif self.detecting == CavityProperty.RTGOUY:
            if owner.direction == "x":
                self.target = &cav_ws.rt_gouy_x
            else:
                self.target = &cav_ws.rt_gouy_y
        elif self.detecting == CavityProperty.MODESEP:
            if owner.direction == "x":
                self.target = &cav_ws.Df_x
            else:
                self.target = &cav_ws.Df_y
        elif self.detecting == CavityProperty.STABILITY:
            if owner.direction == "x":
                self.target = &cav_ws.gx
            else:
                self.target = &cav_ws.gy
        elif self.detecting == CavityProperty.RESOLUTION:
            if owner.direction == "x":
                self.target = &cav_ws.Sx
            else:
                self.target = &cav_ws.Sy
        else:
            raise RuntimeError(f"Invalid value for CavityProperty: {self.detecting}")

        # None of the properties detectable by this workspace require
        # valid beam traces so mark it as ok to compute outputs for
        # invalid trace data points
        self.ignore_sim_mask = True

        self.set_output_fn(cp_detector_output)

cp_detector_output = OutputFuncWrapper.make_from_ptr(c_cp_detector_output)
cdef c_cp_detector_output(DetectorWorkspace dws):
    cdef CPDetectorWorkspace ws = <CPDetectorWorkspace> dws

    # Convert to degrees first if detecting round-trip Gouy phase
    if ws.detecting == CavityProperty.RTGOUY:
        return degrees(ws.target[0])

    return ws.target[0]


cdef class CPDetectorABCDWorkspace(DetectorWorkspace):
    """Workspace for obtaining cavity round-trip ABCD matrix, used
    by :class:`.CavityPropertyDetector`."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        cdef CavityWorkspace cav_ws = self.sim.cavity_workspaces[owner.cavity]
        if owner.direction == "x":
            self.abcd = cav_ws.abcd_x
        else:
            self.abcd = cav_ws.abcd_y

        self.ignore_sim_mask = True

        self.set_output_fn(cp_detector_abcd_output)

cp_detector_abcd_output = OutputFuncWrapper.make_from_ptr(c_cp_detector_abcd_output)
cdef c_cp_detector_abcd_output(DetectorWorkspace dws):
    cdef CPDetectorABCDWorkspace ws = <CPDetectorABCDWorkspace> dws

    return ws.abcd.base.copy()


cdef class CPDetectorModeWorkspace(DetectorWorkspace):
    """Workspace for cavity property calculations, used by :class:`.CavityPropertyDetector`,
    for properties which rely on the cavity eigenmode."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        self.detecting = owner.detecting

        cdef CavityWorkspace cav_ws = self.sim.cavity_workspaces[owner.cavity]
        cdef Py_ssize_t n_id = self.sim.carrier.node_id(owner.cavity.source)
        # Set the q ptr to cav_ws eigenmode of relevant plane
        if owner.direction == "x":
            self.q = &self.sim.trace[n_id].qx
            self.is_stable = &cav_ws.is_stable_x
        else:
            self.q = &self.sim.trace[n_id].qy
            self.is_stable = &cav_ws.is_stable_y

        self.is_changing = self.sim.trace[n_id].is_changing

        if self.detecting == CavityProperty.EIGENMODE:
            self.compute_func = NULL
            self.initial = -1.0
            self.set_output_fn(cp_detector_mode_output)
        else:
            if self.detecting == CavityProperty.SOURCE_SIZE:
                self.compute_func = bp_beamsize
            elif self.detecting == CavityProperty.SOURCE_WAISTSIZE:
                self.compute_func = bp_waistsize
            elif self.detecting == CavityProperty.SOURCE_DISTANCE:
                self.compute_func = bp_waistpos
            elif self.detecting == CavityProperty.SOURCE_RAYLEIGH:
                self.compute_func = bp_rayleigh
            elif self.detecting == CavityProperty.SOURCE_DIVERGENCE:
                self.compute_func = bp_divergence
            elif self.detecting == CavityProperty.SOURCE_ROC:
                self.compute_func = bp_radius_curvature
            elif self.detecting == CavityProperty.SOURCE_DEFOCUS:
                self.compute_func = bp_defocus
            else:
                raise RuntimeError(f"Invalid value for CavityProperty: {self.detecting}")

            if self.is_stable[0]:
                self.initial = self.compute_func(self.q)
            else:
                self.initial = np.nan

            self.set_output_fn(cp_detector_from_mode_output)

cp_detector_from_mode_output = OutputFuncWrapper.make_from_ptr(c_cp_detector_from_mode_output)
cdef c_cp_detector_from_mode_output(DetectorWorkspace dws):
    cdef CPDetectorModeWorkspace ws = <CPDetectorModeWorkspace> dws

    if not ws.is_stable[0]:
        return np.nan

    if not ws.is_changing:
        return ws.initial

    return ws.compute_func(ws.q)

cp_detector_mode_output = OutputFuncWrapper.make_from_ptr(c_cp_detector_mode_output)
cdef c_cp_detector_mode_output(DetectorWorkspace dws):
    cdef CPDetectorModeWorkspace ws = <CPDetectorModeWorkspace> dws

    if not ws.is_stable[0]:
        return np.nan

    if ws.q_as_bp:
        return BeamParam(ws.q.wavelength, ws.q.nr, ws.q.q)

    return ws.q.q


### Astigmatism Detector ###


cdef class AstigmatismDetectorWorkspace(DetectorWorkspace):
    """Workspace for node astigmatism calculations used by :class:`.AstigmatismDetector`."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        cdef Py_ssize_t n_id = self.sim.carrier.node_id(owner.node)
        self.q = &self.sim.trace[n_id]

        self.initial = 1.0 - bp_overlap(&self.q.qx, &self.q.qy)

        self.set_output_fn(astig_detector_output)

astig_detector_output = OutputFuncWrapper.make_from_ptr(c_astig_detector_output)
cdef c_astig_detector_output(DetectorWorkspace dws):
    cdef AstigmatismDetectorWorkspace ws = <AstigmatismDetectorWorkspace> dws

    if not ws.q.is_changing:
        return ws.initial

    return 1.0 - bp_overlap(&ws.q.qx, &ws.q.qy)


### Mode Mismatch Detector ###


cdef class ModeMismatchDetectorWorkspace(DetectorWorkspace):
    """Workspace for mode mismatch (at a node coupling) calculations
    used by :class:`.ModeMismatchDetector`."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        node1, node2 = owner.in_node, owner.out_node
        comp = node1.component
        direction = owner.direction

        cdef ConnectorWorkspace ws = None
        for cws in self.sim.workspaces:
            if cws.owner is comp:
                ws = cws
                break

        if ws is None:
            raise RuntimeError(
                "Bug encountered! Could not find a workspace associated "
                f"with the component of name {comp.name}"
            )

        cdef Py_ssize_t n1_id = self.sim.carrier.node_id(node1)
        cdef Py_ssize_t n2_id = self.sim.carrier.node_id(node2)
        if owner.direction == "x":
            self.q1 = &self.sim.trace[n1_id].qx
            self.q2 = &self.sim.trace[n2_id].qx
        else:
            self.q1 = &self.sim.trace[n1_id].qy
            self.q2 = &self.sim.trace[n2_id].qy

        self.pscale = 100 if owner.in_percent else 1

        p1name, p2name = node1.port.name, node2.port.name
        # First case is if connector ws has single abcd
        # matrix view for all couplings
        if hasattr(ws, "abcd"):
            self.abcd = ws.abcd
        # Next is a connector ws with abcd views same for
        # each node coupling but different for x,y planes
        # e.g. a lens
        elif hasattr(ws, f"abcd_{direction}"):
            self.abcd = getattr(ws, f"abcd_{direction}")
        # If not then check for connector ws with abcd views
        # which are different for each node coupling and x,y
        # plane --- e.g. a mirror or beam splitter
        elif hasattr(ws, f"abcd_{p1name}{p2name}_{direction}"):
            self.abcd = getattr(ws, f"abcd_{p1name}{p2name}_{direction}")
        # Finally, if the connector does not define any abcd
        # matrix views then simply use identity matrix
        else:
            self.abcd = np.eye(2)

        # Determine whether the node coupling is one with changing mismatch
        self.is_changing = (node1, node2) in self.sim.mismatched_node_couplings
        cdef complex_t q1_matched = transform_q(self.abcd, self.q1.q, self.q1.nr, self.q2.nr)
        self.initial = self.pscale * (1.0 - overlap(q1_matched, self.q2.q))

        self.set_output_fn(mode_mismatch_detector_output)

mode_mismatch_detector_output = OutputFuncWrapper.make_from_ptr(c_mode_mismatch_detector_output)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef c_mode_mismatch_detector_output(DetectorWorkspace dws):
    cdef ModeMismatchDetectorWorkspace ws = <ModeMismatchDetectorWorkspace> dws

    if not ws.is_changing:
        return ws.initial

    cdef complex_t q1_matched = transform_q(ws.abcd, ws.q1.q, ws.q1.nr, ws.q2.nr)
    return ws.pscale * (1.0 - overlap(q1_matched, ws.q2.q))


### Accumulated Gouy Phase Detector ###


cdef class GouyDetectorWorkspace(DetectorWorkspace):
    """Workspace for accumulated Gouy phase calculations used by :class:`.Gouy`."""
    def __init__(self, owner, sim):
        super().__init__(owner, sim)

        cdef list spaces = owner.spaces
        self.Nspaces = len(spaces)

        self.gouy_targets = <double**> calloc(self.Nspaces, sizeof(double*))
        if self.gouy_targets == NULL:
            raise MemoryError()

        cdef SpaceWorkspace sws
        cdef Py_ssize_t i, j
        for i in range(self.Nspaces):
            space = spaces[i]
            for ws in sim.workspaces:
                if ws.owner is space:
                    sws = <SpaceWorkspace> ws
                    if owner.direction == "x":
                        self.gouy_targets[i] = &(sws.sv.gouy_x)
                    else:
                        self.gouy_targets[i] = &(sws.sv.gouy_y)

                    break

            if self.gouy_targets[i] == NULL:
                raise RuntimeError(
                    "Bug encountered! Could not find a workspace associated "
                    f"with the Space of name {space.name}"
                )

        self.set_output_fn(acc_gouy_detector_output)

    def __dealloc__(self):
        if self.gouy_targets != NULL:
            free(self.gouy_targets)
            self.gouy_targets = NULL

acc_gouy_detector_output = OutputFuncWrapper.make_from_ptr(c_acc_gouy_detector_output)
cdef c_acc_gouy_detector_output(DetectorWorkspace dws):
    cdef:
        GouyDetectorWorkspace ws = <GouyDetectorWorkspace> dws

        Py_ssize_t i
        double acc = 0.0

    for i in range(ws.Nspaces):
        acc += ws.gouy_targets[i][0]

    return acc



#cython: boundscheck=False, wraparound=False, initializedcheck=False

r"""
Functions for computing images of a beam at arbitrary points
in the interferometer configuration.

.. _camera_equations:

Camera Equations
----------------

Each function in this sub-module has two modes - single frequency (mimicking
amplitude detectors at each coordinate) and multi frequency (mimicking CCD
cameras).

.. rubric:: Single-frequency mode

If the argument `f` (i.e. the field frequency to probe) is specified then this
function computes the amplitude and phase of the light field at this given
frequency, for the specified `x` and `y` coordinate. The light field at frequency
:math:`\omega_{\mathrm{i}}` is given by a complex number (:math:`z`) and is calculated
as follows:

.. math::
    z(x, y) = \displaystyle\sum_{\mathrm{j}} \sum_{nm} u_{nm}(x, y) a_{\mathrm{j}nm}
    \quad\text{with}\quad
    \left\{\,\mathrm{j}\,|\,\mathrm{j} \in [0, \dots, N] \wedge \omega_{\mathrm{j}} = \omega_{\mathrm{i}}\right\}.

.. rubric:: Multi-frequency mode

Otherwise, if `f` is not specified, then this function acts like a CCD camera for the
given pixel. It plots the *beam intensity* as a function of the `x` and `y` coordinates
given. The output is a real number computed as:

.. math::
    s(x, y) = \displaystyle\sum_{\mathrm{ij}} \sum_{nm} u_{nm}(x, y) u_{nm}^*(x,y) a_{\mathrm{i}nm} a_{\mathrm{j}nm}^*
    \quad\text{with}\quad
    \left\{\,\mathrm{i,j}\,|\,\mathrm{i,j} \in [0, \dots, N] \wedge \omega_{\mathrm{i}} = \omega_{\mathrm{j}}\right\}.
"""

import logging

from cython.parallel import prange
cimport numpy as np
import numpy as np

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, creal
from finesse.cymath.complex cimport crotate2
from finesse.cymath.gaussbeam cimport bp_waistsize
from finesse.cymath.math cimport cos, sin, ceil, floor
from finesse.cymath.math cimport float_eq, nmin, nmax
from finesse.cymath.homs cimport (
    unm_ws_recache_from_bp,
    unm_factor_store_init,
    unm_factor_store_recache,
    unm_factor_store_free,
    u_nm__fast
)

from finesse.utilities.cyomp cimport determine_nthreads_even

ctypedef (double*,) ptr_tuple_1

cdef extern from "constants.h":
    double INV_ROOT2
    double complex COMPLEX_0


LOGGER = logging.getLogger(__name__)


cdef class CameraWorkspace(MaskedDetectorWorkspace):
    """Workspace class for cameras."""

    def __init__(self, owner, sim, values=None):
        super().__init__(owner, sim, values)

        self.node_id = self.sim.carrier.node_id(owner.node)
        self.q = &self.sim.trace[self.node_id]

        x = owner.xdata
        y = owner.ydata
        if owner.w0_scaled:
            x *= bp_waistsize(&self.q.qx)
            y *= bp_waistsize(&self.q.qy)

        if isinstance(x, np.ndarray):
            self.x_view = x
            self.scan_ax = XAXIS
        else:
            self.x_view = np.array([x], dtype=np.float64)

        if isinstance(y, np.ndarray):
            self.y_view = y
            self.scan_ax = YAXIS
        else:
            self.y_view = np.array([y], dtype=np.float64)

        self.xpts = self.x_view.shape[0]
        self.ypts = self.y_view.shape[0]

        cdef int outer_pts = nmax(self.xpts, self.ypts)
        # Nominal number of threads will be outer pixel loop size // 50
        self.nthreads = determine_nthreads_even(outer_pts, 50)

        LOGGER.info(
            "Using %d threads for %s outer pixel loop.",
            self.nthreads,
            self.owner.name,
        )

        self.phase_cache = np.zeros((self.num_unmasked_HOMs, 2))

        self.cache(initial=True)

    def __dealloc__(self):
        unm_factor_store_free(&self.ufs)

    cpdef cache(self, bint initial=False):
        """Cache, or re-cache, the u_nm variables and Gouy phase data."""
        cdef:
            Py_ssize_t i
            Py_ssize_t p # mode index
            int n, m

            double n0 = 0.0
            double m0 = 0.0
            double phase

            int max_n, max_m

        unm_ws_recache_from_bp(&self.uws, &self.q.qx, &self.q.qy)
        if initial:
            max_n = self.sim.model_data.max_n
            max_m = self.sim.model_data.max_m
            unm_factor_store_init(&self.ufs, &self.uws, max_n, max_m)
        else:
            unm_factor_store_recache(&self.ufs, &self.uws)

        if not self.sim.model_data.zero_tem00_gouy:
            n0 = 0.5
            m0 = 0.5

        for i in range(self.num_unmasked_HOMs):
            p = self.unmasked_mode_indices[i]
            n = self.sim.model_data.homs_view[p][0]
            m = self.sim.model_data.homs_view[p][1]

            phase = (n + n0) * self.uws.gouyx + (m + m0) * self.uws.gouyy

            self.phase_cache[i][0] = cos(phase)
            self.phase_cache[i][1] = -sin(phase)

### CCD type camera workspaces ###

cdef class CCDWorkspace(CameraWorkspace):
    def __init__(self, owner, sim, out):
        super().__init__(owner, sim)

        self.out = out


cdef class CCDLineWorkspace(CameraWorkspace):
    def __init__(self, owner, sim, out):
        super().__init__(owner, sim)

        self.out = out

### Field / complex camera type workspaces ###

cdef class ComplexCameraValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_1 ptr = (&self.f, )
        cdef tuple params = ("f", )
        self.setup(params, sizeof(ptr), <double**>&ptr)

cdef class ComplexCameraWorkspace(CameraWorkspace):
    def __init__(self, owner, sim):
        super().__init__(owner, sim, ComplexCameraValues())
        self.v = <ComplexCameraValues>self.values

cdef class FieldCameraWorkspace(ComplexCameraWorkspace):
    def __init__(self, owner, sim, out):
        super().__init__(owner, sim)

        self.out = out

cdef class FieldLineWorkspace(ComplexCameraWorkspace):
    def __init__(self, owner, sim, out):
        super().__init__(owner, sim)

        self.out = out


cdef complex_t field_beam_pixel(ComplexCameraWorkspace cws, double x, double y) nogil:
    cdef:
        Py_ssize_t freq_idx, k, field_idx
        int n, m
        complex_t z_ij = COMPLEX_0
        complex_t unm, at
        double cph, sph
        frequency_info_t *freq

        Py_ssize_t node_id = cws.node_id

        Py_ssize_t Nfreqs = cws.sim.carrier.optical_frequencies.size
        Py_ssize_t Nfields = cws.num_unmasked_HOMs

    for freq_idx in range(Nfreqs):
        freq = &cws.sim.carrier.optical_frequencies.frequency_info[freq_idx]
        if float_eq(freq.f, cws.v.f):
            at = COMPLEX_0
            for k in range(Nfields):
                field_idx = cws.unmasked_mode_indices[k]
                n = cws.sim.model_data.homs_view[field_idx][0]
                m = cws.sim.model_data.homs_view[field_idx][1]

                cph = cws.phase_cache[k][0]
                sph = cws.phase_cache[k][1]
                unm = u_nm__fast(&cws.uws, &cws.ufs, n, m, x, y)
                at += (
                    unm
                    * crotate2(
                        cws.sim.carrier.get_out_fast(node_id, freq_idx, field_idx),
                        cph,
                        sph
                    )
                    * INV_ROOT2
                )

            z_ij += at

    return z_ij


cdef double ccd_beam_pixel(CameraWorkspace cws, double x, double y) nogil:
    cdef:
        Py_ssize_t freq_idx_outer, freq_idx_inner, k, field_idx
        int n, m
        complex_t z_ij = COMPLEX_0
        complex_t unm, at1, at2
        double cph, sph

        Py_ssize_t node_id = cws.node_id

        Py_ssize_t Nfreqs = cws.sim.carrier.optical_frequencies.size
        Py_ssize_t Nfields = cws.num_unmasked_HOMs

    for freq_idx_outer in range(Nfreqs):
        for freq_idx_inner in range(Nfreqs):
            if not float_eq(
                cws.sim.carrier.optical_frequencies.frequency_info[freq_idx_outer].f,
                cws.sim.carrier.optical_frequencies.frequency_info[freq_idx_inner].f
            ):
                continue

            at1 = at2 = COMPLEX_0

            for k in range(Nfields):
                field_idx = cws.unmasked_mode_indices[k]
                n = cws.sim.model_data.homs_view[field_idx][0]
                m = cws.sim.model_data.homs_view[field_idx][1]

                cph = cws.phase_cache[k][0]
                sph = cws.phase_cache[k][1]
                unm = u_nm__fast(&cws.uws, &cws.ufs, n, m, x, y)
                at1 += (
                    unm
                    * crotate2(
                        cws.sim.carrier.get_out_fast(node_id, freq_idx_outer, field_idx),
                        cph,
                        sph
                    )
                    * INV_ROOT2
                )
                at2 += (
                    unm
                    * crotate2(
                        cws.sim.carrier.get_out_fast(node_id, freq_idx_inner, field_idx),
                        cph,
                        sph
                    )
                    * INV_ROOT2
                )

            z_ij += at1 * conj(at2)

    return creal(z_ij)


field_pixel_output = OutputFuncWrapper.make_from_ptr(c_field_pixel_output)
cdef c_field_pixel_output(DetectorWorkspace dws):
    cdef:
        FieldPixelWorkspace ws = <CameraWorkspace> dws

    if ws.q.is_changing:
        ws.cache()

    return field_beam_pixel(ws, ws.x_view[0], ws.y_view[0])


ccd_pixel_output = OutputFuncWrapper.make_from_ptr(c_ccd_pixel_output)
cdef c_ccd_pixel_output(DetectorWorkspace dws):
    cdef:
        CameraWorkspace ws = <CameraWorkspace> dws

    if ws.q.is_changing:
        ws.cache()

    return ccd_beam_pixel(ws, ws.x_view[0], ws.y_view[0])


field_line_output = OutputFuncWrapper.make_from_ptr(c_field_line_output)
cdef c_field_line_output(DetectorWorkspace dws):
    cdef:
        FieldLineWorkspace ws = <CameraWorkspace> dws
        Py_ssize_t i, Npts
        double x, y

        int nthreads = ws.nthreads


    if ws.q.is_changing:
        ws.cache()

    if ws.scan_ax == XAXIS:
        Npts = ws.xpts
    else:
        Npts = ws.ypts

    if ws.scan_ax == XAXIS:
        y = ws.y_view[0]
        for i in prange(Npts, nogil=True, num_threads=nthreads, schedule="static"):
            ws.out[i] = field_beam_pixel(ws, ws.x_view[i], y)
    else:
        x = ws.x_view[0]
        for i in prange(Npts, nogil=True, num_threads=nthreads, schedule="static"):
            ws.out[i] = field_beam_pixel(ws, x, ws.y_view[i])

    return ws.out.base.copy()

ccd_line_output = OutputFuncWrapper.make_from_ptr(c_ccd_line_output)
cdef c_ccd_line_output(DetectorWorkspace dws):
    cdef:
        CCDLineWorkspace ws = <CameraWorkspace> dws
        Py_ssize_t i, Npts
        double x, y

        int nthreads = ws.nthreads

    if ws.q.is_changing:
        ws.cache()

    if ws.scan_ax == XAXIS:
        Npts = ws.xpts
    else:
        Npts = ws.ypts

    if ws.scan_ax == XAXIS:
        y = ws.y_view[0]
        for i in prange(Npts, nogil=True, num_threads=nthreads, schedule="static"):
            ws.out[i] = ccd_beam_pixel(ws, ws.x_view[i], y)
    else:
        x = ws.x_view[0]
        for i in prange(Npts, nogil=True, num_threads=nthreads, schedule="static"):
            ws.out[i] = ccd_beam_pixel(ws, x, ws.y_view[i])

    return ws.out.base.copy()

field_camera_output = OutputFuncWrapper.make_from_ptr(c_field_camera_output)
cdef c_field_camera_output(DetectorWorkspace dws):
    cdef:
        FieldCameraWorkspace ws = <CameraWorkspace> dws
        Py_ssize_t i, j

        int nthreads = ws.nthreads

    if ws.q.is_changing:
        ws.cache()

    for i in prange(ws.xpts, nogil=True, num_threads=nthreads, schedule="static"):
        for j in range(ws.ypts):
            ws.out[i][j] = field_beam_pixel(ws, ws.x_view[i], ws.y_view[j])

    return ws.out.base.copy()

ccd_output = OutputFuncWrapper.make_from_ptr(c_ccd_output)
cdef c_ccd_output(DetectorWorkspace dws):
    cdef:
        CCDWorkspace ws = <CameraWorkspace> dws
        Py_ssize_t i, j

        int nthreads = ws.nthreads

    if ws.q.is_changing:
        ws.cache()

    for i in prange(ws.xpts, nogil=True, num_threads=nthreads, schedule="static"):
        for j in range(ws.ypts):
            ws.out[i][j] = ccd_beam_pixel(ws, ws.x_view[i], ws.y_view[j])

    return ws.out.base.copy()



# distutils: sources = src/finesse/tinyexpr.c
# distutils: include_dirs = src/finesse/

"""Compiled symbolic expressions used internally via parameters and element workspaces.

This sub-module only exposes C code so can only be used by other Cython
extensions. The symbolic expression struct ``cy_expr`` is used in workspaces
(see :class:`.ElementWorkspace`) and parameter code (see :class:`.Parameter`)
for quick evaluation of changing symbolic expressions.

The ``cy_expr`` struct, and associated functions, are wrappers around the C
based math parsing and evaluation engine, `tinyexpr <https://github.com/codeplea/tinyexpr>`_.
"""

from libc.stdlib cimport malloc, free


cdef cy_expr* cy_expr_new() nogil:
    cdef cy_expr* ce_p = <cy_expr*> malloc(sizeof(cy_expr))
    ce_p.expr = NULL
    ce_p.variables = NULL
    return ce_p

cdef int cy_expr_init(cy_expr* ex, object operation) except -1:
    """Initialise the te_expr and te_variable type fields using the
    operation object (should be an instance of Operation)."""
    cdef str op_str = str(operation)
    cdef list params = operation.parameters() # get the dependent parameter-refs
    cdef int Nparams = len(params)

    ex.variables = <te_variable*> malloc(Nparams * sizeof(te_variable))
    if not ex.variables:
        raise MemoryError()

    cdef Py_ssize_t i
    cdef Parameter p
    for i in range(Nparams):
        pref = params[i] # the ParameterRef object

        # Replace the corresponding parameter name in the operation string with
        # the tinyexpr compatible name version (see ParameterRef.cyexpr_name)
        op_str = op_str.replace(pref.name, pref.cyexpr_name.decode())
        # Also need to replace "quantity**n" with "quantity^n" as te expects powers in this form
        op_str = op_str.replace("**", "^")

        p = pref.parameter
        ex.variables[i] = te_variable(pref.cyexpr_name, &p.__cvalue, 0, NULL)

    cdef int err # position of parsing error in operation expression
    # Make byte str of operation expression, and store it in
    # cy_expr instance for info / debugging purposes later
    byte_op_str = op_str.encode("UTF-8")
    ex.expression = byte_op_str
    ex.expr = te_compile(byte_op_str, ex.variables, Nparams, &err)

    if not ex.expr:
        error_loc_str = "    {m: <{pos}}^".format(m='', pos=str(err - 1))
        raise RuntimeError(
            "Bug encountered! Internal cy_expr parsing error on:\n"
           f"    {op_str}\n" +
            error_loc_str +
           f"\nError near here"
        )

    return 0

cdef void cy_expr_free(cy_expr* ex) nogil:
    if ex != NULL:
        if ex.expr != NULL:
            te_free(ex.expr)
        if ex.variables != NULL:
            free(ex.variables)

        free(ex)
        ex = NULL

cdef double cy_expr_eval(cy_expr* ex) nogil:
    """Evaluate the cythonised symbolic expression."""
    return te_eval(ex.expr)



from finesse.cmatrix cimport SubCCSView, SubCCSView2DArray
from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport cexp, crotate, conj
from finesse.cymath.math cimport radians, cosh, sinh
from finesse.simulations.base cimport frequency_info_t, NodeBeamParam
from finesse.simulations.basematrix cimport MatrixSystemSolver, CarrierSignalMatrixSimulation
import numpy as np
cimport numpy as np
cimport cython

ctypedef (double*, double*, double*) ptr_tuple_3

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD
    double DB2R

cdef class SqueezerConnections:
    def __cinit__(self):
        pass

cdef class SqueezerNoiseSources:
    def __cinit__(self, MatrixSystemSolver mtx):
        cdef:
            int Nfo =  mtx.optical_frequencies.size

        self.P1o = SubCCSView2DArray(Nfo, Nfo)
        self.ptrs.P1o = <PyObject***>self.P1o.views

cdef class SqueezerValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_3 ptr = (&self.db, &self.angle, &self.f)
        cdef tuple params = ("db","angle","f")
        self.setup(params, sizeof(ptr), <double**>&ptr)

cdef class SqueezerWorkspace(ConnectorWorkspace):
    def __init__(self, object owner, CarrierSignalMatrixSimulation sim, bint refill):
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            values=SqueezerValues(),
            noise_sources=SqueezerNoiseSources(sim.signal) if sim.signal else None
        )
        self.v = self.values
        self.ns = self.signal.noise_sources if sim.signal else None
        self.qn_coeffs = np.zeros((sim.model_data.num_HOMs, sim.model_data.num_HOMs), dtype=np.complex128)
        self.qn_coeffs_diag = np.zeros(sim.model_data.num_HOMs, dtype=np.complex128)


squeezer_fill_qnoise = FillFuncWrapper.make_from_ptr(c_squeezer_fill_qnoise)
cdef object c_squeezer_fill_qnoise(ConnectorWorkspace cws):
    r"""
    Fills the quantum noise right hand side (RHS) vector corresponding
    to the squeezed-light source `squeezer`.
    """
    cdef:
        SqueezerWorkspace ws = cws

        # Laser quantum noise injection
        complex_t n = ws.sim.model_data.UNIT_VACUUM / 2
        complex_t phs = crotate(1, 2 * radians(ws.v.angle))
        double r = ws.v.db * DB2R
        complex_t qn
        squeezer_noise_sources noises = ws.ns.ptrs
        frequency_info_t *ifreq
        frequency_info_t *ofreq

        Py_ssize_t i

    # TODO: Shouldn't this quantum noise be frequency-dependent, as for other noise sources?
    ws.qn_coeffs_diag[:] = n

    for i in range(ws.sim.signal.optical_frequencies.size):
        ifreq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        if ifreq.audio_carrier_index != ws.fsrc_car_idx:
            (<SubCCSView>noises.P1o[ifreq.index][ifreq.index]).fill_za(n)
            continue
        for j in range(ws.sim.signal.optical_frequencies.size):
            ofreq = &(ws.sim.signal.optical_frequencies.frequency_info[j])
            if ofreq.audio_carrier_index != ws.fsrc_car_idx:
                continue
            # Reflections

            if ws.sim.signal.optical_frequencies.frequency_info[i].audio_order > 0:
                if i == j:
                    qn = n * cosh(2 * r)
                else:
                    qn = n * sinh(2 * r) * phs
            else:
                if i == j:
                    qn = conj(n * cosh(2 * r))
                else:
                    qn = conj(n * sinh(2 * r) * phs)

            # We only want to squeeze the main mode of the interferometer, so just set the first
            # element of the relevant matrix/diagonal
            if i == j:
                ws.qn_coeffs_diag[0] = qn
                (<SubCCSView>noises.P1o[ifreq.index][ofreq.index]).fill_zd(ws.qn_coeffs_diag)
            else:
                ws.qn_coeffs[0][0] = qn
                (<SubCCSView>noises.P1o[ifreq.index][ofreq.index]).fill_zm(ws.qn_coeffs)


squeezer_fill_rhs = FillFuncWrapper.make_from_ptr(c_squeezer_fill_rhs)
cdef object c_squeezer_fill_rhs(ConnectorWorkspace cws):
    cdef:
        SqueezerWorkspace ws = <SqueezerWorkspace>cws

    if not ws.sim.is_modal:
        ws.sim.set_source_fast(
            ws.node_id, ws.fsrc_car_idx, 0, 0,
        )
    else:
        for i in range(ws.signal.nhoms):
            ws.sim.set_source_fast(
                ws.node_id, ws.fsrc_car_idx, i, 0,
            )



from finesse.components.workspace import Connections
from finesse.cymath.complex cimport conj, cexp, complex_t

ctypedef (double*, double*) ptr_tuple_2

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD


cdef class SignalGeneratorValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_2 ptr = (&self.amplitude, &self.phase)
        cdef tuple params = ("amplitude","phase")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class SignalGeneratorWorkspace(ConnectorWorkspace):
    def __init__(self, owner, sim, refill):
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            Connections(),
            Connections(),
            SignalGeneratorValues()
        )
        self.v = <SignalGeneratorValues>self.values


siggen_fill_rhs = FillFuncWrapper.make_from_ptr(c_siggen_fill_rhs)
cdef object c_siggen_fill_rhs(ConnectorWorkspace cws):
    cdef:
        SignalGeneratorWorkspace ws = <SignalGeneratorWorkspace>cws
        complex_t A

    A = (ws.v.amplitude * ws.scaling) * cexp(1j * ws.v.phase * DEG2RAD)
    ws.sim.signal.set_source_fast_2(ws.rhs_idx, A)



#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""Fast computations of Higher-Order Mode related properties.

This module provides functions for computing properties of HOMs, such as
the spatial distribution :math:`u_{nm}(x,y,z;q_x,q_y)` of Hermite-Gauss
modes.

.. note::
    These are low-level functions intended for use, by developers, in Cython
    extensions. To calculate HG mode profiles, for example, users should
    instead use the Python class :class:`.HGMode` defined in :mod:`finesse.gaussian`.
"""

from libc.stdlib cimport calloc, free

cimport numpy as np
import numpy as np

from cython.parallel import prange

from finesse.cymath.complex cimport complex_t, inverse_unsafe, ceq, cpow_re, cexp
from finesse.cymath.gaussbeam cimport waist_size, beam_size, gouy
from finesse.cymath.math cimport ceil, floor
from finesse.cymath.math cimport factorial, sqrt_factorial, hermite, nmin

from finesse.utilities.cyomp cimport determine_nthreads_even


cdef extern from "complex.h" nogil:
    double cimag(double complex z)
    double creal(double complex z)
    double complex csqrt(double complex z)
    double complex conj(double complex z)

cdef extern from "math.h" nogil:
    double exp(double arg)
    double sqrt(double arg)

cdef extern from "constants.h":
    long double PI
    double TWO_ON_PI_QRT # (2/pi)^{1/4}
    double ROOT2
    double complex COMPLEX_0


### Spatial distribution (u_m, u_m) functions and workspace for HG modes ###

# See Eq. (9.34) in "Interferometer Techniques for Gravitational Wave Detection"
# as main point of reference for equations used in functions below

## General workspace struct for u_nm calculations ##

cdef void unm_ws_recache_from_bp(
    unm_workspace* uws,
    const beam_param* qx,
    const beam_param* qy,
    int* n=NULL,
    int* m=NULL,
): # Initialise uws from beam_param constructs
    unm_ws_recache(uws, qx.q, qy.q, qx.nr, qx.wavelength, n, m)

cdef void unm_ws_recache(
    unm_workspace* uws,
    complex_t qx,
    complex_t qy,
    double nr,
    double lambda0,
    int* n=NULL,
    int* m=NULL,
) nogil:
    cdef:
        double zrx = cimag(qx)
        double zry

    uws.k = 2.0 * PI / lambda0 * nr
    uws.root_w0x = sqrt(waist_size(qx, nr, lambda0))
    uws.gouyx = gouy(qx)
    uws.root2_on_wx = ROOT2 / beam_size(qx, nr, lambda0)
    uws.inv_qx = inverse_unsafe(qx)
    uws.xgauss_exponent = -1j * 0.5 * uws.k * uws.inv_qx
    uws.z1x = csqrt((0.0 + zrx * 1j) / qx)
    uws.z2x = csqrt((0.0 + zrx * 1j) * conj(qx) / ((0.0 - zrx * 1j) * qx))
    # If n is given then store it and calculate the terms
    # before Hn(x') present in the u_n equation
    if n != NULL:
        uws.n = n[0]
        uws.xpre_factor = (
            TWO_ON_PI_QRT
            * 1 / (sqrt(2**uws.n) * sqrt_factorial(uws.n) * uws.root_w0x)
            * uws.z1x * cpow_re(uws.z2x, uws.n)
        )
    else:
        uws.n = -1
        uws.xpre_factor = COMPLEX_0

    # Equality check quicker than other calculations
    if ceq(qx, qy):
        uws.root_w0y = uws.root_w0x
        uws.gouyy = uws.gouyx
        uws.root2_on_wy = uws.root2_on_wx
        uws.inv_qy = uws.inv_qx
        uws.ygauss_exponent = uws.xgauss_exponent
        uws.z1y = uws.z1x
        uws.z2y = uws.z2x
    else:
        zry = cimag(qy)
        uws.root_w0y = sqrt(waist_size(qy, nr, lambda0))
        uws.gouyy = gouy(qy)
        uws.root2_on_wy = ROOT2 / beam_size(qy, nr, lambda0)
        uws.inv_qy = inverse_unsafe(qy)
        uws.ygauss_exponent = -1j * 0.5 * uws.k * uws.inv_qy
        uws.z1y = csqrt((0.0 + zry * 1j) / qy)
        uws.z2y = csqrt((0.0 + zry * 1j) * conj(qy) / ((0.0 - zry * 1j) * qy))

    # If m is given then store it and calculate the terms
    # before Hm(x') present in the u_m equation
    if m != NULL:
        uws.m = m[0]
        uws.ypre_factor = (
            TWO_ON_PI_QRT
            * 1 / (sqrt(2**uws.m) * sqrt_factorial(uws.m) * uws.root_w0y)
            * uws.z1y * cpow_re(uws.z2y, uws.m)
        )
    else:
        uws.m = -1
        uws.ypre_factor = COMPLEX_0

## Additional specialised u_nm workspace struct holding pre-factor power caches ##

cdef void unm_factor_store_init(
    unm_factor_store* ufs,
    const unm_workspace* uws,
    int max_n,
    int max_m,
) nogil:
    ufs.nsize = 1 + max_n
    ufs.msize = 1 + max_m
    ufs.xpre_factor_cache = <complex_t*> calloc(ufs.nsize, sizeof(complex_t))
    ufs.ypre_factor_cache = <complex_t*> calloc(ufs.msize, sizeof(complex_t))

    unm_factor_store_recache(ufs, uws)

cdef void unm_factor_store_recache(unm_factor_store* ufs, const unm_workspace* uws) nogil:
    cdef complex_t x_const = TWO_ON_PI_QRT * uws.z1x / uws.root_w0x
    cdef complex_t y_const = TWO_ON_PI_QRT * uws.z1y / uws.root_w0y

    cdef int n
    for n in range(ufs.nsize):
        ufs.xpre_factor_cache[n] = (
            x_const * cpow_re(uws.z2x, n) * 1 / (sqrt(2**n) * sqrt_factorial(n))
        )

    cdef int m
    for m in range(ufs.msize):
        ufs.ypre_factor_cache[m] = (
            y_const * cpow_re(uws.z2y, m) * 1 / (sqrt(2**m) * sqrt_factorial(m))
        )

cdef void unm_factor_store_free(unm_factor_store* ufs) nogil:
    if ufs.xpre_factor_cache != NULL: free(ufs.xpre_factor_cache)
    if ufs.ypre_factor_cache != NULL: free(ufs.ypre_factor_cache)

## The u_n, u_m and u_nm functions themselves ##

cdef complex_t u_nm(const unm_workspace* uws, int n, int m, double x, double y) nogil:
    return u_n(uws, n, x) * u_m(uws, m, y)

cdef complex_t u_nm__fast(
    const unm_workspace* uws,
    const unm_factor_store* ufs,
    int n, int m,
    double x, double y
) nogil:
    return u_n__fast(uws, ufs, n, x) * u_m__fast(uws, ufs, m, y)

cdef complex_t u_nmconst(const unm_workspace* uws, double x, double y) nogil:
    return u_nconst(uws, x) * u_mconst(uws, y)


cdef complex_t u_n(const unm_workspace* uws, int n, double x) nogil:
    return (
        TWO_ON_PI_QRT
        * 1 / (sqrt(2**n) * sqrt_factorial(n) * uws.root_w0x)
        * uws.z1x * cpow_re(uws.z2x, n)
        * hermite(n, uws.root2_on_wx * x)
        * cexp(uws.xgauss_exponent * x * x)
    )

cdef complex_t u_n__fast(
    const unm_workspace* uws,
    const unm_factor_store* ufs,
    int n,
    double x,
) nogil:
    return (
        ufs.xpre_factor_cache[n]
        * hermite(n, uws.root2_on_wx * x)
        * cexp(uws.xgauss_exponent * x * x)
    )

cdef complex_t u_nconst(const unm_workspace* uws, double x) nogil:
    return (
        uws.xpre_factor
        * hermite(uws.n, uws.root2_on_wx * x)
        * cexp(uws.xgauss_exponent * x * x)
    )


cdef complex_t u_m(const unm_workspace* uws, int m, double y) nogil:
    return (
        TWO_ON_PI_QRT
        * 1 / (sqrt(2**m) * sqrt_factorial(m) * uws.root_w0y)
        * uws.z1y * cpow_re(uws.z2y, m)
        * hermite(m, uws.root2_on_wy * y)
        * cexp(uws.ygauss_exponent * y * y)
    )

cdef complex_t u_m__fast(
    const unm_workspace* uws,
    const unm_factor_store* ufs,
    int m,
    double y,
) nogil:
    return (
        ufs.ypre_factor_cache[m]
        * hermite(m, uws.root2_on_wy * y)
        * cexp(uws.ygauss_exponent * y * y)
    )

cdef complex_t u_mconst(const unm_workspace* uws, double y) nogil:
    return (
        uws.ypre_factor
        * hermite(uws.m, uws.root2_on_wy * y)
        * cexp(uws.ygauss_exponent * y * y)
    )


### Misc. functions ###


cpdef Py_ssize_t field_index(int n, int m, const int[:, ::1] homs) nogil:
    cdef:
        Py_ssize_t i
        Py_ssize_t N = homs.shape[0]
        int ni, mi

    for i in range(N):
        ni = homs[i][0]
        mi = homs[i][1]

        if ni == n and mi == m:
            return i

    # if mode not found return size of homs array
    return N


cpdef bint in_mask(int n, int m, const int[:, ::1] mask) nogil:
    cdef:
        Py_ssize_t i
        Py_ssize_t Nmasks = mask.shape[0]
        int n_mask, m_mask

    for i in range(Nmasks):
        n_mask = mask[i][0]
        m_mask = mask[i][1]

        if n_mask == n and m_mask == m:
            return True

    return False


cdef class HGModeWorkspace:
    """Fast computation of Hermite-Gauss spatial distributions.

    This workspace class is used internally by :class:`.HGMode`. Users should
    only ever interact with the :class:`.HGMode` object rather than this class."""
    def __init__(
        self, int n, int m, complex_t qx, complex_t qy, double nr, double lambda0
    ):
        self.n = n
        self.m = m

        self.set_values(qx, qy, nr, lambda0)

    cpdef set_values(self, qx=None, qy=None, nr=None, lambda0=None):
        cdef:
            complex_t qx_, qy_
            double nr_, lambda0_

        if qx is not None:
            self.qx = complex(qx)

        if qy is not None:
            self.qy = complex(qy)

        self.is_astigmatic = not ceq(self.qx, self.qy)

        if nr is not None:
            self.nr = float(nr)

        if lambda0 is not None:
            self.lambda0 = float(lambda0)

        unm_ws_recache(&self.uws, self.qx, self.qy, self.nr, self.lambda0, &self.n, &self.m)

    cpdef u_n(self, x, complex_t[::1] out=None):
        from finesse.utilities.misc import is_iterable

        if not is_iterable(x):
            return u_nconst(&self.uws, float(x))

        cdef double[::1] xs = x
        cdef Py_ssize_t N = xs.shape[0]

        cdef np.ndarray[complex_t, ndim=1] out_arr
        if out is None:
            out_arr = np.zeros(N, dtype=np.complex128)
            out = out_arr

        cdef int nthreads = determine_nthreads_even(N, 50)

        cdef Py_ssize_t i
        for i in prange(N, nogil=True, num_threads=nthreads):
            out[i] = u_nconst(&self.uws, xs[i])

        return out.base

    cpdef u_m(self, y, complex_t[::1] out=None):
        from finesse.utilities.misc import is_iterable

        if not is_iterable(y):
            return u_mconst(&self.uws, float(y))

        cdef double[::1] ys = y
        cdef Py_ssize_t N = ys.shape[0]

        cdef np.ndarray[complex_t, ndim=1] out_arr
        if out is None:
            out_arr = np.zeros(N, dtype=np.complex128)
            out = out_arr

        cdef int nthreads = determine_nthreads_even(N, 50)

        cdef Py_ssize_t i
        for i in prange(N, nogil=True, num_threads=nthreads):
            out[i] = u_mconst(&self.uws, ys[i])

        return out.base

    cpdef u_nm(self, x, y, out=None):
        U_n = self.u_n(x)
        if self.n != self.m or self.is_astigmatic:
            U_m = self.u_m(y)
        else:
            U_m = U_n

        # NumPy outer product much faster than manually
        # computing u_nm so use this here instead
        U_nm = np.outer(U_n, U_m)
        if out is not None:
            out[:] = U_nm
            return out
        else:
            return U_nm



#cython: boundscheck=False, wraparound=False, initializedcheck=False

from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, cexp
from finesse.cymath.math cimport exp, sqrt, cos, sin, asin, isfinite
from finesse.cymath.math cimport radians
from finesse.simulations.base cimport NodeBeamParam
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation, MatrixSystemSolver
from finesse.frequency cimport frequency_info_t
from finesse.utilities import refractive_index
from finesse.cmatrix cimport SubCCSView
from libc.stdlib cimport calloc, free, malloc, realloc
from libc.string cimport memset
from numpy cimport ndarray

import numpy as np
cimport numpy as np
from scipy.special import jn

import logging

ctypedef (double*, double*, double*) ptr_tuple_3

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD

LOGGER = logging.getLogger(__name__)

cdef double[:,::1] abcd_unity = np.eye(2, dtype=float)

cdef class ModulatorValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_3 ptr = (&self.f, &self.midx, &self.phase)
        cdef tuple params = ("f", "midx", "phase")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class ModulatorOpticalConnections:
    """Contains C accessible references to submatrices for
    optical connections for this element.
    """
    def __cinit__(self, MatrixSystemSolver mtx):
        cdef int Nf = mtx.optical_frequencies.size

        self.P1i_P2o = SubCCSView2DArray(Nf, Nf)
        self.P2i_P1o = SubCCSView2DArray(Nf, Nf)
        self.opt_conn_ptrs.P1i_P2o = <PyObject***>self.P1i_P2o.views
        self.opt_conn_ptrs.P2i_P1o = <PyObject***>self.P2i_P1o.views


cdef class ModulatorSignalConnections:
    """Contains C accessible references to submatrices for
    optical connections for this element.
    """
    def __cinit__(self, MatrixSystemSolver mtx):
        cdef int Nf = mtx.optical_frequencies.size

        self.amp_P2o = SubCCSView2DArray(1, Nf)
        self.amp_P1o = SubCCSView2DArray(1, Nf)
        self.phs_P2o = SubCCSView2DArray(1, Nf)
        self.phs_P1o = SubCCSView2DArray(1, Nf)

        self.sig_conn_ptrs.amp_P2o = <PyObject***>self.amp_P2o.views
        self.sig_conn_ptrs.amp_P1o = <PyObject***>self.amp_P1o.views
        self.sig_conn_ptrs.phs_P2o = <PyObject***>self.phs_P2o.views
        self.sig_conn_ptrs.phs_P1o = <PyObject***>self.phs_P1o.views


cdef void set_coupling_orders(ndarray orders, int set_order, double f, mtx, modulator_coupling_order** rtn, Py_ssize_t* N, dict coupling_orders):
    """Returns a [frequency input index, frequency output index, order] array
    free must be called on rtn* once finished with object.
    """
    for f1 in mtx.optical_frequencies.frequencies:
        coupling_orders[f1, f1] = 0
        for f2 in mtx.optical_frequencies.frequencies:
            df = f2.f - f1.f
            order = df / float(f)
            if hasattr(order, "eval"):
                order = order.eval()

            if not order.is_integer():
                continue

            order = int(round(order))

            if abs(order) <= set_order:
                coupling_orders[f1, f2] = order
                coupling_orders[f2, f1] = -order

    items = tuple(coupling_orders.items())
    N[0] = len(items)
    rtn[0] = <modulator_coupling_order*>calloc(N[0], sizeof(modulator_coupling_order))

    for i, ((f1, f2), _order) in enumerate(items):
        rtn[0][i].f1_index = f1.index
        rtn[0][i].f2_index = f2.index
        rtn[0][i].order = _order
        rtn[0][i].order_index = int(np.argwhere(orders == _order))


cdef class ModulatorWorkspace(KnmConnectorWorkspace):
    def __init__(self, object owner, CarrierSignalMatrixSimulation sim, bint refill):
        cdef Py_ssize_t i
        super().__init__(owner,
            sim, refill, refill,
            ModulatorOpticalConnections(sim.carrier),
            ModulatorSignalConnections(sim.signal) if sim.signal else None,
            ModulatorValues()
        )

        self.nr1 = refractive_index(owner.p1)
        self.nr2 = refractive_index(owner.p2)

        self.cvalues = self.values
        self.mcc = self.carrier.connections

        if sim.signal:
            self.mcs = self.signal.connections
            self.amp_signal_enabled = owner.amp.i.full_name in sim.signal.nodes
            self.phs_signal_enabled = owner.phs.i.full_name in sim.signal.nodes
        else:
            self.mcs = None

        if sim.is_modal:
            self.set_knm_info(
                "P1i_P2o", abcd_x=abcd_unity, abcd_y=abcd_unity, nr_from=self.nr1, nr_to=self.nr2, is_transmission=True
            )
            self.set_knm_info(
                "P2i_P1o", abcd_x=abcd_unity, abcd_y=abcd_unity, nr_from=self.nr2, nr_to=self.nr1, is_transmission=True
            )

        order = int(owner.order.eval())
        self.N_orders = 2 * order + 1
        self.orders = np.zeros(self.N_orders, dtype=np.int32)
        self.phases = np.zeros(self.N_orders, dtype=np.complex128)
        self.amps = np.zeros(self.N_orders, dtype=np.float64)

        for i in range(-order, order + 1):
            if i < 0:
                self.orders[self.N_orders + i] = i
            else:
                self.orders[i] = i

        if self.values.mod_type == ModulatorType.am:
            if self.amp_signal_enabled:
                raise NotImplementedError("amplitude signal injection not supported yet with amplitude modulators")
            if self.phs_signal_enabled:
                raise NotImplementedError("phase signal injection not supported yet with amplitude modulators")

        self.factors_12 = np.zeros((self.N_orders, sim.model_data.num_HOMs, sim.model_data.num_HOMs), dtype=np.complex128)
        self.factors_21 = np.zeros((self.N_orders, sim.model_data.num_HOMs, sim.model_data.num_HOMs), dtype=np.complex128)

        self.carrier_frequency_couplings = {}
        self.signal_frequency_couplings = {}

        # need to free this memory later
        set_coupling_orders(
            self.orders,
            order, owner.f,
            sim.carrier, &self.car_coupling_orders,
            &self.N_car_coupling_orders,
            self.carrier_frequency_couplings
        )
        if sim.signal:
            set_coupling_orders(
                self.orders,
                order, owner.f, sim.signal, &self.sig_coupling_orders,
                &self.N_sig_coupling_orders,
                self.signal_frequency_couplings
            )
            self.increment_sig = <bint*>calloc(self.sim.signal.optical_frequencies.size, sizeof(bint))
            self.eye = np.eye(sim.model_data.num_HOMs, dtype=np.complex128)

    def __dealloc__(self):
        if self.sig_coupling_orders:
            free(self.sig_coupling_orders)
        if self.car_coupling_orders:
            free(self.car_coupling_orders)
        if self.increment_sig:
            free(self.increment_sig)

    cpdef fill_quantum_matrix(self):
        cdef:
            modulator_coupling_order* mco = NULL
            modulator_optical_connections *conn = &self.mcs.opt_conn_ptrs
            Py_ssize_t i

        for i in range(self.N_sig_coupling_orders):
            mco = (self.sig_coupling_orders + i)
            if mco.f1_index != mco.f2_index:
                if conn.P1i_P2o[mco.f1_index][mco.f2_index]:
                    (<SubCCSView>conn.P1i_P2o[mco.f1_index][mco.f2_index]).fill_negative_za(0)
                if conn.P2i_P1o[mco.f1_index][mco.f2_index]:
                    (<SubCCSView>conn.P2i_P1o[mco.f1_index][mco.f2_index]).fill_negative_za(0)
            else:
                if conn.P1i_P2o[mco.f1_index][mco.f2_index]:
                    (<SubCCSView>conn.P1i_P2o[mco.f1_index][mco.f2_index]).fill_negative_zm(self.eye)
                if conn.P2i_P1o[mco.f1_index][mco.f2_index]:
                    (<SubCCSView>conn.P2i_P1o[mco.f1_index][mco.f2_index]).fill_negative_zm(self.eye)


cdef object compute_factors(ModulatorWorkspace ws):
    """
    ToDo
    ----
    This function still needs optimising, plenty of Python calls still
    """
    cdef:
        double phi = radians(ws.cvalues.phase)
        double midx = ws.cvalues.midx
        Py_ssize_t i

    if ws.cvalues.mod_type == ModulatorType.am:
        ws.amps[:] = 0.25 * midx
        ws.amps[0] = 1.0 - 0.5 * midx
    elif ws.cvalues.mod_type == ModulatorType.pm:
        ws.amps[:] = jn(ws.orders, midx)
        phi += 0.5 * PI

    np.exp(1j * phi * ws.orders, out=ws.phases)

    if ws.cvalues.positive_only:
        ws.amps[0] = 1.0 - 0.5 * (1.0 - ws.amps_view[0])

        for i in range(-ws.cvalues.order, 0):
            ws.amps[ws.N_orders + i] = 0.0

    # TODO ddb can use a faster outer product here probably
    np.multiply.outer(ws.amps * ws.phases, ws.K12.data, out=ws.factors_12)
    np.multiply.outer(ws.amps * ws.phases, ws.K21.data, out=ws.factors_21)


cdef void fill_optical_2_optical(
    ModulatorWorkspace ws,
    modulator_optical_connections *conn,
    modulator_coupling_order* cpl_orders,
    Py_ssize_t N_cpl_orders
):
    cdef modulator_coupling_order* mco = NULL
    cdef Py_ssize_t i = 0

    assert(cpl_orders)
    assert(conn)
    compute_factors(ws)

    for i in range(N_cpl_orders):
        mco = (cpl_orders + i)
        if conn.P1i_P2o[mco.f1_index][mco.f2_index]:
            (<SubCCSView>conn.P1i_P2o[mco.f1_index][mco.f2_index]).fill_za_zm(
                -1, ws.factors_12[mco.order_index][:]
            )
        if conn.P2i_P1o[mco.f1_index][mco.f2_index]:
            (<SubCCSView>conn.P2i_P1o[mco.f1_index][mco.f2_index]).fill_za_zm(
                -1, ws.factors_21[mco.order_index][:]
            )


modulator_carrier_fill = FillFuncWrapper.make_from_ptr(c_modulator_carrier_fill)
cdef object c_modulator_carrier_fill(ConnectorWorkspace cws):
    cdef ModulatorWorkspace ws = <ModulatorWorkspace>cws
    fill_optical_2_optical(ws, &ws.mcc.opt_conn_ptrs, ws.car_coupling_orders, ws.N_car_coupling_orders)


modulator_signal_optical_fill = FillFuncWrapper.make_from_ptr(c_modulator_signal_optical_fill)
cdef object c_modulator_signal_optical_fill(ConnectorWorkspace cws):
    cdef ModulatorWorkspace ws = <ModulatorWorkspace>cws
    fill_optical_2_optical(ws, &ws.mcs.opt_conn_ptrs, ws.sig_coupling_orders, ws.N_sig_coupling_orders)


modulator_signal_phase_fill = FillFuncWrapper.make_from_ptr(c_modulator_signal_phase_fill)
cdef object c_modulator_signal_phase_fill(ConnectorWorkspace cws):
    cdef ModulatorWorkspace ws = <ModulatorWorkspace>cws
    if ws.mod_type == ModulatorType.pm:
        fill_phase_mod_amp_2_optical_signals(ws)


modulator_signal_amp_fill = FillFuncWrapper.make_from_ptr(c_modulator_signal_amp_fill)
cdef object c_modulator_signal_amp_fill(ConnectorWorkspace cws):
    cdef ModulatorWorkspace ws = <ModulatorWorkspace>cws
    if ws.mod_type == ModulatorType.pm:
        fill_phase_mod_phs_2_optical_signals(ws)


cdef object fill_phase_mod_phs_2_optical_signals(
        ModulatorWorkspace ws
    ):
    """Fills the phase noise to output optical field coupling for a phase modulator
    """
    cdef:
        double midx = ws.cvalues.midx
        modulator_coupling_order* mco = NULL
        Py_ssize_t i = 0
        int car_cpl_index
        frequency_info_t *f
        frequency_info_t* fs[2]
        complex_t factor
        double k
        modulator_signal_connections *signal_conn = &ws.mcs.sig_conn_ptrs
        modulator_optical_connections *carrier_conn = &ws.mcc.opt_conn_ptrs
    # we use this flag array to keep track of whether we will
    # increment or reset a coupling.
    memset(ws.increment_sig, 0, ws.sim.signal.optical_frequencies.size * sizeof(bint))
    # Steps here:
    # 1 - Get each carrier coupling order
    # 2 - compute the factor i*k*midx/2
    # 3 - get the upper and lower sidebands around the outgoing carrier frequency
    # 4 - apply this phase modulation only around incoming carrier that gets coupled
    for i in range(ws.N_car_coupling_orders):
        mco = (ws.car_coupling_orders + i)
        # get audio sidebands
        f = &ws.sim.carrier.optical_frequencies.frequency_info[mco.f2_index]
        k = ws.sim.model_data.k0 * (1 + f.f_car[0]/ws.sim.model_data.f0)
        fs[0] = &ws.sim.signal.optical_frequencies.frequency_info[f.audio_upper_index]
        fs[1] = &ws.sim.signal.optical_frequencies.frequency_info[f.audio_lower_index]
        factor = 1j * k * midx * 0.5 # two half factor from two cosine gains

        for f in fs:
            if signal_conn.phs_P2o[0][f.index]:
                (<SubCCSView>signal_conn.phs_P2o[0][f.index]).fill_prop_za (
                    (<SubCCSView>carrier_conn.P1i_P2o[mco.f1_index][mco.f2_index]), 0, factor, ws.increment_sig[f.index]
                )
            if signal_conn.phs_P1o[0][f.index]:
                (<SubCCSView>signal_conn.phs_P1o[0][f.index]).fill_prop_za (
                    (<SubCCSView>carrier_conn.P2i_P1o[mco.f1_index][mco.f2_index]), 0, factor, ws.increment_sig[f.index]
                )
            # flag that any other coupling into this output frequency needs to increment now
            # otherwise we'll overwrite previouslly set couplings
            ws.increment_sig[f.index] |= 1


cdef object fill_phase_mod_amp_2_optical_signals(
        ModulatorWorkspace ws
    ):
    """Fills amplitude noise input to optical output coupling elements.
    Based on a small relative modulation index excitation
    .. math::
        J_{k}\\left(m \\left(\\delta \\cos{\\left(\\Omega t \\right)} + 1\\right)\\right)
    .. math::
        \\delta \\left(\\left(\\frac{m J_{k - 1}\\left(m\\right)}{4} - \\frac{m J_{k + 1}\\left(m\\right)}{4}\\right) e^{i \\Omega t} + \\left(\\frac{m J_{k - 1}\\left(m\\right)}{4} - \\frac{m J_{k + 1}\\left(m\\right)}{4}\\right) e^{- i \\Omega t}\\right) + J_{k}\\left(m\\right)
    """
    cdef:
        double midx = ws.cvalues.midx
        modulator_coupling_order* mco = NULL
        Py_ssize_t i = 0
        int car_cpl_index
        double J_m1, J, J_p1
        frequency_info_t *f
        frequency_info_t* fs[2]
        complex_t factor
        modulator_signal_connections *signal_conn = &ws.mcs.sig_conn_ptrs
        modulator_optical_connections *carrier_conn = &ws.mcc.opt_conn_ptrs
    # we use this flag array to keep track of whether we will
    # increment or reset a coupling.
    memset(ws.increment_sig, 0, ws.sim.signal.optical_frequencies.size * sizeof(bint))

    # Steps here:
    # 1 - Get each carrier coupling order
    # 2 - compute the bessel function delta*(J(order-1)-J(order+1)) factor
    # 3 - get the upper and lower sidebands around the outgoing carrier frequency
    # 4 - apply this amplitude modulation only around incoming carrier that gets coupled
    for i in range(ws.N_car_coupling_orders):
        mco = (ws.car_coupling_orders + i)
        # need to divide by J(order, midx) because this is included in the carrier coupling
        # and not in the signal sideband terms
        J = jn(mco.order, midx)
        J_m1 = jn(mco.order-1, midx)
        J_p1 = jn(mco.order+1, midx)
        # get audio sidebands
        f = &ws.sim.carrier.optical_frequencies.frequency_info[mco.f2_index]
        fs[0] = &ws.sim.signal.optical_frequencies.frequency_info[f.audio_upper_index]
        fs[1] = &ws.sim.signal.optical_frequencies.frequency_info[f.audio_lower_index]
        factor = midx * 0.5 * 0.5 * (J_m1 - J_p1)/J # two half factor from two cosine gains

        for f in fs:
            if signal_conn.amp_P2o[0][f.index]:
                (<SubCCSView>signal_conn.amp_P2o[0][f.index]).fill_prop_za (
                    (<SubCCSView>carrier_conn.P1i_P2o[mco.f1_index][mco.f2_index]), 0, factor, ws.increment_sig[f.index]
                )
            if signal_conn.amp_P1o[0][f.index]:
                (<SubCCSView>signal_conn.amp_P1o[0][f.index]).fill_prop_za (
                    (<SubCCSView>carrier_conn.P2i_P1o[mco.f1_index][mco.f2_index]), 0, factor, ws.increment_sig[f.index]
                )
            # flag that any other coupling into this output frequency needs to increment now
            # otherwise we'll overwrite previouslly set couplings
            ws.increment_sig[f.index] |= 1


# def modulator_fill(object modulator, ModulatorWorkspace ws, MatrixSystemSolver mtx, connections, dict values, dict cporders, unicode
#         mod_type, bint quantum_fill=False):


#     for (f1, f2), _order in cporders.items():
#         if quantum_fill:
#             if f1.index == f2.index:
#                 with mtx.component_edge_fill3(ws.owner_id, connections.P1i_P2o_idx, f1.index, f2.index) as mat:
#                     mat[:] = 1
#                 with mtx.component_edge_fill3(ws.owner_id, connections.P2i_P1o_idx, f1.index, f2.index) as mat:
#                     mat[:] = 1
#             else:
#                 with mtx.component_edge_fill3(ws.owner_id, connections.P1i_P2o_idx, f1.index, f2.index) as mat:
#                     mat[:] = 0
#                 with mtx.component_edge_fill3(ws.owner_id, connections.P2i_P1o_idx, f1.index, f2.index) as mat:
#                     mat[:] = 0
#         else:
#             with mtx.component_edge_fill3(ws.owner_id, connections.P1i_P2o_idx, f1.index, f2.index) as mat:
#                 mat[:] = factors_12[_order][:]
#             with mtx.component_edge_fill3(ws.owner_id, connections.P2i_P1o_idx, f1.index, f2.index) as mat:
#                 mat[:] = factors_21[_order][:]



#!/usr/bin/env python
# pylint: disable = invalid-name, missing-docstring

"""Check classes are documented properly as per numpydoc requirements.

This currently checks:
  - constructor parameters are documented in the class docstring, not __init__
  - listed sections in docstrings are valid

Note: because this script imports finesse modules directly, it should be executed within the same
environment used to develop Finesse. In pre-commit, for example, this means setting the script
language as "system" and not "python".

Author: Sean Leavey
"""

import sys
import argparse
import inspect
from pathlib import Path
from importlib import import_module
import warnings
from numpydoc.docscrape import FunctionDoc, ClassDoc
import finesse


FINESSE_ROOT = Path(finesse.__file__.replace("__init__.py", ""))


def check_module(path):
    path = Path(path).resolve()

    try:
        module_path = path.relative_to(FINESSE_ROOT)
    except ValueError:
        # Specified file is not part of the finesse package.
        return 0

    module_name = "." + str(module_path.stem)
    package = str(finesse.__name__ + "." + str(module_path.parent).replace("/", "."))
    package = package.rstrip(".")

    try:
        module = import_module(module_name, package)
    except ModuleNotFoundError:
        return 0

    has_issue = False

    for _, obj in inspect.getmembers(module):
        if not inspect.isclass(obj):
            continue

        if obj.__module__ != module.__name__:
            # Reject imported modules.
            continue

        clsobj = f"{module_path}::{obj.__name__}"
        initmeth = None

        try:
            if obj.__init__.__module__ == obj.__module__:
                # There is an init method.
                initmeth = obj.__init__
        except AttributeError:
            # Probably a C parent.
            pass

        # Parse docstring.
        initdoc = None
        with warnings.catch_warnings(record=True) as warnlist:
            # Import classdoc so that it triggers warnings caught by the context manager.
            try:
                ClassDoc(obj)

                if initmeth:
                    initdoc = FunctionDoc(initmeth)
            except Exception as e:
                print(f"error while processing docstring for {obj}: {e}")

                has_issue = True

            if warnlist:
                for wrng in warnlist:
                    print(f"{clsobj}: numpydoc warning: {wrng.message}")

                has_issue = True

        if initdoc and initdoc.get("Parameters"):
            print(
                f"{clsobj}: constructor parameters should be documented in the class "
                "docstring, not __init__ (see https://finesse.readthedocs.io/en/latest/developer/documenting.html#writing-sphinx-compatible-docstrings)."
            )

            has_issue = True

    return 1 if has_issue else 0


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("filenames", nargs="*")
    args = parser.parse_args(sys.argv)

    retv = 0

    for filename in args.filenames:
        retv |= check_module(filename)

    sys.exit(retv)



from finesse.components.workspace cimport ConnectorWorkspace
from finesse.knm cimport knm_bh_workspace
from finesse.simulations.base cimport NodeBeamParam
from finesse.simulations.basematrix cimport CarrierSignalMatrixSimulation
from finesse.cymath.complex cimport complex_t, DenseZMatrix
from cpython.ref cimport PyObject
from finesse.parameter cimport Parameter
import numpy as np
cimport numpy as np

cdef struct KnmInfo:
    bint has_been_set # False if component has not set the info for it
    int from_port_idx
    int to_port_idx
    bint is_transmission
    # Refractive indices of adjacent spaces
    double nr_from
    double nr_to
    # C-contiguous ABCD matrix views for this connection: assumes it is a 2x2!!
    double* abcd_x
    double* abcd_y
    # references to the Knm workspaces this information is associated with
    knm_bh_workspace *K_ws_x
    knm_bh_workspace *K_ws_y
    # misalignment information
    double *beta_x
    double *beta_y
    double *alpha
    double beta_y_factor
    double beta_x_factor
    bint alpha_is_changing
    bint beta_x_is_changing
    bint beta_y_is_changing
    # Pointer to a KnmMatrix
    DenseZMatrix *mtx
    # Pointer to bayerhelms scatter matrix function to use
    # -> see end of set_knm_info definition for details
    void (*bhelms_mtx_func)(
        const knm_bh_workspace*,
        const knm_bh_workspace*,
        const int*,
        complex_t*,
        Py_ssize_t,
        int,
    )
    # Pointer to buffer of knm loss data
    double* loss


cdef class KnmConnectorWorkspace(ConnectorWorkspace):
    cdef:
        readonly dict o2o # dict[str, (node, node)] for optical connections
        Py_ssize_t* onode_ids # An array of input, output ordered node ids for each port in index order
        knm_bh_workspace* Kws # Array of workspaces for calculating Knm matricies
        double[::1] total_losses  # Temporary array used during loss-induced quantum noise calculations
        int N_opt_ports
        int N_opt_conns
        KnmInfo *oconn_info

    cdef initialise_knm_workspaces(KnmConnectorWorkspace self)
    cdef void free_knm_workspaces(KnmConnectorWorkspace self) nogil
    cdef void flag_changing_knm_workspaces(KnmConnectorWorkspace self)
    cdef void update_changing_knm_workspaces(KnmConnectorWorkspace self) nogil
    cdef void compute_scattering_matrices(KnmConnectorWorkspace self)
    cdef void compute_knm_losses(KnmConnectorWorkspace self) nogil

    cpdef set_knm_info(
        self,
        connection,
        double[:,::1] abcd_x = ?,
        double[:,::1] abcd_y = ?,
        double nr_from=?, double nr_to=?,
        bint is_transmission=?,
        Parameter beta_x = ?,
        double beta_x_factor = ?,
        Parameter beta_y = ?,
        double beta_y_factor = ?,
        Parameter alpha = ?
    )
